///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2009, 2010 Sebastian Pancratz                               //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

/**
 * \file    fmpz_poly_q-test.c
 * \brief   Test file for code coverage in fmpz_poly_q.c and fmpz_poly_q.h
 * \author  Sebastian Pancratz
 * \date    Oct 2009 -- Jan 2010
 * 
 * The code in this file aims to cover the code provided by the rational 
 * function field implementation in the file <tt>fmpz_poly_q.c</tt> and 
 * <tt>fmpz_poly_q.h</tt>.
 * 
 * For almost all functions that are available through the header file 
 * <tt>fmpz_poly_q.h</tt>, there is a corresponding test function in this 
 * file.
 * 
 * Typically, for a function taking \f$k\f$ arguments the testing method has 
 * the signature <tt>int test_foo(char * in1, ..., char * in<i>k</i>, 
 * char * out, int verbose)</tt>.  The function then sets up \f$k\f$ 
 * independent rational functions, computes the output and obtains a string 
 * representation, finally comparing this to the rational function represented 
 * by <tt>out</tt>.  The function returns <tt>0</tt> for a test passed and 
 * <tt>-1</tt> otherwise.  This idea is modified accordingly when the function 
 * which is to be tested takes further arguments of primitive type.  In 
 * addition to this, for some functions that have to be careful about handling 
 * input aliasing correctly, e.g. addition and multiplication, there are 
 * multiple testing functions simulating this.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "fmpz_poly_q.h"

int test_set(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_set(rop, op);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_set: passed\n");
        }
        else
        {
            printf("test_set: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_set_si(long x, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop;
    char * res;
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_set_si(rop, x);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_set_si: passed\n");
        }
        else
        {
            printf("test_set_si: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_swap(char * in1, char * in2, char * out1, char * out2, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    char * res1;
    char * res2;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_swap(op1, op2);
    
    res1 = fmpz_poly_q_to_string(op1);
    res2 = fmpz_poly_q_to_string(op2);
    
    ans = !strcmp(out1, res1) & !strcmp(out2, res2);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_swap: passed\n");
        }
        else
        {
            printf("test_swap: failed\n");
            printf("    Expected \"%s\" \"%s\", got \"%s\" \"%s\"\n", out1, out2, res1, res2);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res1);
    free(res2);
    
    return ans ? 1 : 0;
}

int test_zero(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_zero(op);
    
    res = fmpz_poly_q_to_string(op);
    
    ans = !strcmp(res, out);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_zero: passed\n");
        }
        else
        {
            printf("test_zero: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    free(res);
    
    return ans ? 1 : 0;
}

int test_neg(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_neg(rop, op);
    
    res = fmpz_poly_q_to_string(rop);
    
    ans = !strcmp(res, out);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_neg: passed\n");
        }
        else
        {
            printf("test_neg: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_inv(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
     
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    fmpz_poly_q_init(rop);
    fmpz_poly_q_inv(rop, op);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(res, out);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_inv: passed\n");
        }
        else
        {
            printf("test_inv: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_inv_inplace(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop;
    char * res;
     
    fmpz_poly_q_init(rop);
    fmpz_poly_q_from_string(rop, in);
    fmpz_poly_q_inv(rop, rop);
    
    res = fmpz_poly_q_to_string(rop);
    
    ans = !strcmp(res, out);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_inv_inplace: passed\n");
        }
        else
        {
            printf("test_inv_inplace: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_is_zero(char * in, int out, int verbose)
{
    int ans;
    fmpz_poly_q_t op;
    int res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    res = fmpz_poly_q_is_zero(op);
    
    ans = (res == out);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_equal: passed\n");
        }
        else
        {
            printf("test_equal: failed\n");
            printf("    Expected \"%d\", got \"%d\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    
    return ans ? 1 : 0;
}

int test_is_one(char * in, int out, int verbose)
{
    int ans;
    fmpz_poly_q_t op;
    int res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    res = fmpz_poly_q_is_one(op);
    
    ans = (res == out);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_equal: passed\n");
        }
        else
        {
            printf("test_equal: failed\n");
            printf("    Expected \"%d\", got \"%d\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    
    return ans ? 1 : 0;
}

int test_equal(char * in1, char * in2, int out, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    int res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    res = fmpz_poly_q_equal(op1, op2);
    
    ans = (res == out);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_equal: passed\n");
        }
        else
        {
            printf("test_equal: failed\n");
            printf("    Expected \"%d\", got \"%d\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    
    return ans ? 1 : 0;
}

int test_add(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_add(rop, op1, op2);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    if (verbose)
    {
        if (ans)
        {
            printf("test_add: passed\n");
        }
        else
        {
            printf("test_add: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

// Runs in1 = in1 + in2
int test_add_in_place1(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_add(op1, op1, op2);
    
    res = fmpz_poly_q_to_string(op1);
    ans = !strcmp(out, res);
    if (verbose)
    {
        if (ans)
        {
            printf("test_add_in_place1: passed\n");
        }
        else
        {
            printf("test_add_in_place1: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

// Runs in2 = in1 + in2
int test_add_in_place2(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_add(op2, op1, op2);
    
    res = fmpz_poly_q_to_string(op2);
    ans = !strcmp(out, res);
    if (verbose)
    {
        if (ans)
        {
            printf("test_add_in_place2: passed\n");
        }
        else
        {
            printf("test_add_in_place2: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

// Runs out = in + in
int test_add_in_place3(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_add(rop, op, op);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    if (verbose)
    {
        if (ans)
        {
            printf("test_add_in_place3: passed\n");
        }
        else
        {
            printf("test_add_in_place3: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(rop);
    fmpz_poly_q_clear(op);
    free(res);
    
    return ans ? 1 : 0;
}

int test_sub(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_sub(rop, op1, op2);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_sub: passed\n");
        }
        else
        {
            printf("test_sub: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

// in1 = in1 + in2
int test_sub_in_place1(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_sub(op1, op1, op2);
    
    res = fmpz_poly_q_to_string(op1);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_sub_in_place1: passed\n");
        }
        else
        {
            printf("test_sub_in_place1: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

// in2 = in1 + in2
int test_sub_in_place2(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_sub(op2, op1, op2);
    
    res = fmpz_poly_q_to_string(op2);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_sub_in_place2: passed\n");
        }
        else
        {
            printf("test_sub_in_place2: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

// Runs out = in - in
int test_sub_in_place3(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_sub(rop, op, op);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    if (verbose)
    {
        if (ans)
        {
            printf("test_sub_in_place3: passed\n");
        }
        else
        {
            printf("test_sub_in_place3: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(rop);
    fmpz_poly_q_clear(op);
    free(res);
    
    return ans ? 1 : 0;
}

int test_scalar_mul_si(char * in, long x, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_scalar_mul_si(rop, op, x);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_scalar_mul_si: passed\n");
        }
        else
        {
            printf("test_scalar_mul_si: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_scalar_mul_mpz(char * in, mpz_t x, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_scalar_mul_mpz(rop, op, x);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_scalar_mul_mpz: passed\n");
        }
        else
        {
            printf("test_scalar_mul_mpz: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_scalar_mul_mpq(char * in, mpq_t x, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_scalar_mul_mpq(rop, op, x);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_scalar_mul_mpq: passed\n");
        }
        else
        {
            printf("test_scalar_mul_mpq: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_scalar_div_si(char * in, long x, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_scalar_div_si(rop, op, x);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_scalar_div_si: passed\n");
        }
        else
        {
            printf("test_scalar_div_si: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_scalar_div_mpz(char * in, mpz_t x, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_scalar_div_mpz(rop, op, x);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_scalar_div_mpz: passed\n");
        }
        else
        {
            printf("test_scalar_div_mpz: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_scalar_div_mpq(char * in, mpq_t x, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_scalar_div_mpq(rop, op, x);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_scalar_div_mpq: passed\n");
        }
        else
        {
            printf("test_scalar_div_mpq: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_mul(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_mul(rop, op1, op2);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_mul: passed\n");
        }
        else
        {
            printf("test_mul: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

// in1 = in1 * in2
int test_mul_in_place1(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_mul(op1, op1, op2);
    
    res = fmpz_poly_q_to_string(op1);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_mul_in_place1: passed\n");
        }
        else
        {
            printf("test_mul_in_place1: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

// in2 = in1 * in2
int test_mul_in_place2(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_mul(op2, op1, op2);
    
    res = fmpz_poly_q_to_string(op2);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_mul_in_place2: passed\n");
        }
        else
        {
            printf("test_mul_in_place2: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

// Runs out = in * in
int test_mul_in_place3(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_mul(rop, op, op);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    if (verbose)
    {
        if (ans)
        {
            printf("test_mul_in_place3: passed\n");
        }
        else
        {
            printf("test_mul_in_place3: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(rop);
    fmpz_poly_q_clear(op);
    free(res);
    
    return ans ? 1 : 0;
}

int test_div(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_div(rop, op1, op2);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_div: passed\n");
        }
        else
        {
            printf("test_div: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

// in1 = in1 / in2
int test_div_in_place1(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_div(op1, op1, op2);
    
    res = fmpz_poly_q_to_string(op1);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_div_in_place1: passed\n");
        }
        else
        {
            printf("test_div_in_place1: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

// in2 = in1 / in2
int test_div_in_place2(char * in1, char * in2, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op1, op2;
    char * res;
    
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in1);
    
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in2);
    
    fmpz_poly_q_div(op2, op1, op2);
    
    res = fmpz_poly_q_to_string(op2);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_div_in_place2: passed\n");
        }
        else
        {
            printf("test_div_in_place2: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

// Runs out = in / in
int test_div_in_place3(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_div(rop, op, op);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    if (verbose)
    {
        if (ans)
        {
            printf("test_div_in_place3: passed\n");
        }
        else
        {
            printf("test_div_in_place3: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(rop);
    fmpz_poly_q_clear(op);
    free(res);
    
    return ans ? 1 : 0;
}

int test_power(char * in, unsigned long x, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_power(rop, op, x);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_power: passed\n");
        }
        else
        {
            printf("test_power: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_derivative(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op;
    char * res;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_derivative(rop, op);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_derivative: passed\n");
        }
        else
        {
            printf("test_derivative: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_evaluate(char * in, int numa, int numb, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t op;
    mpq_t rop, a;
    char * res;
    char * str1;
    char * str2;
    int pole;
    
    fmpz_poly_q_init(op);
    fmpz_poly_q_from_string(op, in);
    
    mpq_init(a);
    mpq_set_si(a, numa, numb);
    mpq_init(rop);
    pole = !fmpz_poly_q_evaluate(rop, op, a);
    
    // If we found a pole and the caller suggested this
    if (pole && !strcmp(out, "P"))
    {
        if (verbose)
        {
                printf("test_evaluate: passed\n");
        }
        fmpz_poly_q_clear(op);
        mpq_clear(rop);
        mpq_clear(a);
        return 1;
    }
    if (pole && strcmp(out, "P"))
    {
        if (verbose)
        {
            printf("test_evaluate: failed\n");
            printf("    Expected \"%s\", got a pole\n", out);
        }
        fmpz_poly_q_clear(op);
        mpq_clear(rop);
        mpq_clear(a);
        return 0;
    }
    if (!pole && !strcmp(out, "P"))
    {
        res = mpq_get_str(NULL, 10, rop);
        if (verbose)
        {
            printf("test_evaluate: failed\n");
            printf("    Expected a pole, got \"%s\"\n", res);
        }
        fmpz_poly_q_clear(op);
        mpq_clear(rop);
        mpq_clear(a);
        free(res);
        return 0;
    }
    
    res = mpq_get_str(NULL, 10, rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_evaluate: passed\n");
        }
        else
        {
            printf("test_evaluate: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(op);
    mpq_clear(rop);
    mpq_clear(a);
    free(res);
    
    return ans ? 1 : 0;
}

int test_to_string_pretty(char * in, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop;
    char * res;
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_from_string(rop, in);
    res = fmpz_poly_q_to_string_pretty(rop, "t");
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_to_string_pretty: passed\n");
        }
        else
        {
            printf("test_to_string_pretty: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(rop);
    free(res);
    
    return ans ? 1 : 0;
}

int test_addmul(char * in1, char * in2, char * in3, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op1, op2;
    char * res;
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_from_string(rop, in1);
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in2);
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in3);
    
    fmpz_poly_q_addmul(rop, op1, op2);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_addmul: passed\n");
        }
        else
        {
            printf("test_addmul: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(rop);
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

int test_submul(char * in1, char * in2, char * in3, char * out, int verbose)
{
    int ans;
    fmpz_poly_q_t rop, op1, op2;
    char * res;
    
    fmpz_poly_q_init(rop);
    fmpz_poly_q_from_string(rop, in1);
    fmpz_poly_q_init(op1);
    fmpz_poly_q_from_string(op1, in2);
    fmpz_poly_q_init(op2);
    fmpz_poly_q_from_string(op2, in3);
    
    fmpz_poly_q_submul(rop, op1, op2);
    
    res = fmpz_poly_q_to_string(rop);
    ans = !strcmp(out, res);
    
    if (verbose)
    {
        if (ans)
        {
            printf("test_submul: passed\n");
        }
        else
        {
            printf("test_submul: failed\n");
            printf("    Expected \"%s\", got \"%s\"\n", out, res);
        }
    }
    
    fmpz_poly_q_clear(rop);
    fmpz_poly_q_clear(op1);
    fmpz_poly_q_clear(op2);
    free(res);
    
    return ans ? 1 : 0;
}

int main(int argc, char *argv[])
{
    char * str;
    char * strin;
    char * strout;
    int ans;
    
    fmpz_poly_q_t qpoly1, qpoly2, qpoly3;
    
    fmpz_poly_t zpoly;
    
    mpz_t mpzzero, mpzone, mpztwo;
    mpq_t mpqzero, mpqone, mpqtwo, mpqtwoinv;
    
    //-----------------------------------------------------------------------//
    // Accessing numerator and denominator                                   //
    //-----------------------------------------------------------------------//
    
    fmpz_poly_q_init(qpoly1);
    fmpz_poly_q_from_string(qpoly1, "2  -1 1/2  0 1");
    str = "2  -1 1";
    strout = fmpz_poly_to_string(fmpz_poly_q_numref(qpoly1));
    ans = !strcmp(str, strout);
    if (ans)
    {
        printf("test_numref: passed\n");
    }
    else
    {
        printf("test_numref: failed\n");
        printf("    Expected \"%s\", got \"%s\"\n", str, strout);
    }
    fmpz_poly_q_clear(qpoly1);
    free(strout);
    
    fmpz_poly_q_init(qpoly1);
    fmpz_poly_q_from_string(qpoly1, "2  -1 1/2  0 1");
    str = "2  0 1";
    strout = fmpz_poly_to_string(fmpz_poly_q_denref(qpoly1));
    ans = !strcmp(str, strout);
    if (ans)
    {
        printf("test_denref: passed\n");
    }
    else
    {
        printf("test_denref: failed\n");
        printf("    Expected \"%s\", got \"%s\"\n", str, strout);
    }
    fmpz_poly_q_clear(qpoly1);
    free(strout);
    
    fmpz_poly_q_init(qpoly1);
    fmpz_poly_init(zpoly);
    fmpz_poly_q_from_string(qpoly1, "2  -1 1/2  0 1");
    fmpz_poly_q_get_num(zpoly, qpoly1);
    str = "2  -1 1";
    strout = fmpz_poly_to_string(zpoly);
    ans = !strcmp(str, strout);
    if (ans)
    {
        printf("test_get_num: passed\n");
    }
    else
    {
        printf("test_get_num: failed\n");
        printf("    Expected \"%s\", got \"%s\"\n", str, strout);
    }
    fmpz_poly_q_clear(qpoly1);
    fmpz_poly_clear(zpoly);
    free(strout);
    
    fmpz_poly_q_init(qpoly1);
    fmpz_poly_init(zpoly);
    fmpz_poly_q_from_string(qpoly1, "2  -1 1/2  0 1");
    fmpz_poly_q_get_den(zpoly, qpoly1);
    str = "2  0 1";
    strout = fmpz_poly_to_string(zpoly);
    ans = !strcmp(str, strout);
    if (ans)
    {
        printf("test_get_den: passed\n");
    }
    else
    {
        printf("test_get_den: failed\n");
        printf("    Expected \"%s\", got \"%s\"\n", str, strout);
    }
    fmpz_poly_q_clear(qpoly1);
    fmpz_poly_clear(zpoly);
    free(strout);
    
    fmpz_poly_q_init(qpoly1);
    fmpz_poly_init(zpoly);
    fmpz_poly_q_from_string(qpoly1, "1  1/1  1");
    fmpz_poly_from_string(zpoly, "2  0 1");
    fmpz_poly_q_set_num(qpoly1, zpoly);
    str = "2  0 1";
    strout = fmpz_poly_to_string(fmpz_poly_q_numref(qpoly1));
    ans = !strcmp(str, strout);
    if (ans)
    {
        printf("test_set_num: passed\n");
    }
    else
    {
        printf("test_set_num: failed\n");
        printf("    Expected \"%s\", got \"%s\"\n", str, strout);
    }
    fmpz_poly_q_clear(qpoly1);
    fmpz_poly_clear(zpoly);
    free(strout);
    
    fmpz_poly_q_init(qpoly1);
    fmpz_poly_init(zpoly);
    fmpz_poly_q_from_string(qpoly1, "1  1/1  1");
    fmpz_poly_from_string(zpoly, "2  0 1");
    fmpz_poly_q_set_den(qpoly1, zpoly);
    str = "2  0 1";
    strout = fmpz_poly_to_string(fmpz_poly_q_denref(qpoly1));
    ans = !strcmp(str, strout);
    if (ans)
    {
        printf("test_set_den: passed\n");
    }
    else
    {
        printf("test_set_den: failed\n");
        printf("    Expected \"%s\", got \"%s\"\n", str, strout);
    }
    fmpz_poly_q_clear(qpoly1);
    fmpz_poly_clear(zpoly);
    free(strout);
    
    //-----------------------------------------------------------------------//
    // Canonicalize                                                          //
    //-----------------------------------------------------------------------//
    
    // canonicalize -- mostly tested indirectly
    fmpz_poly_q_init(qpoly1);
    str = "2  -1 1/2  0 1";
    fmpz_poly_q_from_string(qpoly1, str);
    strout = fmpz_poly_q_to_string(qpoly1);
    ans = !strcmp(str, strout);
    if (ans)
    {
        printf("test_canonicalize: passed\n");
    }
    else
    {
        printf("test_canonicalize: failed\n");
        printf("    Expected \"%s\", got \"%s\"\n", str, strout);
    }
    fmpz_poly_q_clear(qpoly1);
    free(strout);
    
    fmpz_poly_q_init(qpoly1);
    str = "2  -1 -1/2  0 1";
    fmpz_poly_q_from_string(qpoly1, "2  1 1/2  0 -1");
    strout = fmpz_poly_q_to_string(qpoly1);
    ans = !strcmp("2  -1 -1/2  0 1", strout);
    if (ans)
    {
        printf("test_canonicalize: passed\n");
    }
    else
    {
        printf("test_canonicalize: failed\n");
        printf("    Expected \"%s\", got \"%s\"\n", str, strout);
    }
    free(strout);
    fmpz_poly_q_clear(qpoly1);
    
    //-----------------------------------------------------------------------//
    // Initialization, memory management and basic operations                //
    //-----------------------------------------------------------------------//
    
    // init -- Tested indirectly
    // clear -- Tested indirectly
    
    // set
    test_set("0", "0", 1);
    test_set("0/1  1", "0", 1);
    test_set("3  -1 0 1/2  0 1", "3  -1 0 1/2  0 1", 1);
    test_set("3  -1 0 1/2  1 1", "2  -1 1", 1);
    
    // set_si
    test_set_si(-1, "1  -1", 1);
    test_set_si(13, "1  13", 1);
    test_set_si(0, "0", 1);
    
    // swap
    test_swap("3  -1 0 1/2  0 1", "1  2/1  3", "1  2/1  3", "3  -1 0 1/2  0 1", 1);
    
    // zero
    test_zero("0", "0", 1);
    test_zero("0/1  1", "0", 1);
    test_zero("3  -1 0 1/2  0 1", "0", 1);
    
    // neg
    test_neg("0", "0", 1);
    test_neg("1  1/1  2", "1  -1/1  2", 1);
    test_neg("3  -1 0 1/2  0 1", "3  1 0 -1/2  0 1", 1);
    
    // inv
    test_inv("1  1/1  2", "1  2", 1);
    test_inv("3  -1 0 1/2  0 1", "2  0 1/3  -1 0 1", 1);
    test_inv("3  -1 0 -1/2  0 1", "2  0 -1/3  1 0 1", 1);
    
    // inv_inplace
    test_inv_inplace("1  1/1  2", "1  2", 1);
    test_inv_inplace("3  -1 0 1/2  0 1", "2  0 1/3  -1 0 1", 1);
    test_inv_inplace("3  -1 0 -1/2  0 1", "2  0 -1/3  1 0 1", 1);
    
    // is_zero
    test_is_zero("0", 1, 1);
    test_is_zero("0/1  1", 1, 1);
    test_is_zero("3  -1 0 1/2  0 1", 0, 1);
    test_is_zero("3  -1 0 1/2  1 1", 0, 1);
    
    // is_one
    test_is_one("0", 0, 1);
    test_is_one("0/1  1", 0, 1);
    test_is_one("1", 0, 1);
    test_is_one("2  1 1/2  1 1", 1, 1);
    test_is_one("3  -1 0 1/2  0 1", 0, 1);
    
    // equal
    test_equal("1  1/1  2", "1  1/1  2", 1, 1);
    test_equal("1  1/1  2", "1  1/1  2", 1, 1);
    test_equal("3  -1 0 1/2  1 1", "2  -1 1", 1, 1);
    test_equal("3  -1 0 1/2  -1 1", "2  -1 1", 0, 1);
    
    //-----------------------------------------------------------------------//
    // Addition and subtraction                                              //
    //-----------------------------------------------------------------------//
    
    // add
    test_add("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "5  1 0 1 0 1/4  0 1 0 1", 1);
    test_add("3  -1 0 1/2  1 1", "1  2/2  -1 1", "3  3 -2 1/2  -1 1", 1);
    test_add("0/2  1 1", "1  2/1  1", "1  2", 1);
    test_add("1  -3/1  4", "0/3  1 0 1", "1  -3/1  4", 1);
    test_add("2  1 1/1  1", "2  -1 1/1  1", "2  0 2", 1);
    test_add("2  1 1/2  0 1", "2  2 1/2  -1 1", "3  -1 2 2/3  0 -1 1", 1);
    test_add("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "4  7 12 7 1/3  2 3 1", 1);
    test_add("2  1 1/2  -1 1", "2  1 1", "3  0 1 1/2  -1 1", 1);
    test_add("1  1/2  1 1", "2  0 1/2  1 1", "1  1", 1);
    test_add("2  1 1/3  4 -4 1", "1  1/2  -2 1", "2  -1 2/3  4 -4 1", 1);
    test_add("3  0 1 1/3  1 2 1", "2  0 -1/2  1 1", "0", 1);
    test_add("2  1 1/2  0 1", "2  -1 1/2  0 1", "1  2", 1);
    test_add("1  1/3  3 5 2", "1  1/3  6 7 2", "1  1/3  2 3 1", 1);
    
    test_add_in_place1("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "5  1 0 1 0 1/4  0 1 0 1", 1);
    test_add_in_place1("3  -1 0 1/2  1 1", "1  2/2  -1 1", "3  3 -2 1/2  -1 1", 1);
    test_add_in_place1("0/2  1 1", "1  2/1  1", "1  2", 1);
    test_add_in_place1("1  -3/1  4", "0/3  1 0 1", "1  -3/1  4", 1);
    test_add_in_place1("2  1 1/1  1", "2  -1 1/1  1", "2  0 2", 1);
    test_add_in_place1("2  1 1/2  0 1", "2  2 1/2  -1 1", "3  -1 2 2/3  0 -1 1", 1);
    test_add_in_place1("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "4  7 12 7 1/3  2 3 1", 1);
    test_add_in_place1("2  1 1/2  -1 1", "2  1 1", "3  0 1 1/2  -1 1", 1);
    test_add_in_place1("1  1/2  1 1", "2  0 1/2  1 1", "1  1", 1);
    test_add_in_place1("2  1 1/3  4 -4 1", "1  1/2  -2 1", "2  -1 2/3  4 -4 1", 1);
    test_add_in_place1("3  0 1 1/3  1 2 1", "2  0 -1/2  1 1", "0", 1);
    test_add_in_place1("2  1 1/2  0 1", "2  -1 1/2  0 1", "1  2", 1);
    
    test_add_in_place2("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "5  1 0 1 0 1/4  0 1 0 1", 1);
    test_add_in_place2("3  -1 0 1/2  1 1", "1  2/2  -1 1", "3  3 -2 1/2  -1 1", 1);
    test_add_in_place2("0/2  1 1", "1  2/1  1", "1  2", 1);
    test_add_in_place2("1  -3/1  4", "0/3  1 0 1", "1  -3/1  4", 1);
    test_add_in_place2("2  1 1/1  1", "2  -1 1/1  1", "2  0 2", 1);
    test_add_in_place2("2  1 1/2  0 1", "2  2 1/2  -1 1", "3  -1 2 2/3  0 -1 1", 1);
    test_add_in_place2("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "4  7 12 7 1/3  2 3 1", 1);
    test_add_in_place2("2  1 1/2  -1 1", "2  1 1", "3  0 1 1/2  -1 1", 1);
    test_add_in_place2("1  1/2  1 1", "2  0 1/2  1 1", "1  1", 1);
    test_add_in_place2("2  1 1/3  4 -4 1", "1  1/2  -2 1", "2  -1 2/3  4 -4 1", 1);
    test_add_in_place2("3  0 1 1/3  1 2 1", "2  0 -1/2  1 1", "0", 1);
    test_add_in_place2("2  1 1/2  0 1", "2  -1 1/2  0 1", "1  2", 1);
    
    test_add_in_place3("2  1 1", "2  2 2", 1);
    test_add_in_place3("2  1 1/1  2", "2  1 1", 1);
    
    // sub
    test_sub("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "5  1 0 3 0 1/4  0 1 0 1", 1);
    test_sub("3  -1 0 1/2  1 1", "1  2/2  -1 1", "3  -1 -2 1/2  -1 1", 1);
    test_sub("0/2  1 1", "1  2/1  1", "1  -2", 1);
    test_sub("1  -3/1  4", "0/3  1 0 1", "1  -3/1  4", 1);
    test_sub("2  1 1/1  1", "2  -1 1/1  1", "1  2", 1);
    test_sub("2  1 1/2  0 1", "2  2 1/2  -1 1", "2  -1 -2/3  0 -1 1", 1);
    test_sub("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "4  -9 -12 -5 -1/3  2 3 1", 1);
    test_sub("2  -1 1/2  0 1", "1  1", "1  -1/2  0 1", 1);
    test_sub("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "5  1 0 3 0 1/4  0 1 0 1", 1);
    test_sub("3  -1 0 1/2  1 1", "1  2/2  -1 1", "3  -1 -2 1/2  -1 1", 1);
    test_sub("0/2  1 1", "1  2/1  1", "1  -2", 1);
    test_sub("1  -3/1  4", "0/3  1 0 1", "1  -3/1  4", 1);
    test_sub("2  1 1/1  1", "2  -1 1/1  1", "1  2", 1);
    test_sub("2  1 1/2  0 1", "2  2 1/2  -1 1", "2  -1 -2/3  0 -1 1", 1);
    test_sub("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "4  -9 -12 -5 -1/3  2 3 1", 1);
    test_sub("2  1 1/2  -1 1", "2  1 1", "3  2 1 -1/2  -1 1", 1);
    test_sub("1  1/2  1 1", "2  0 1/2  1 1", "2  1 -1/2  1 1", 1);
    test_sub("2  1 1/3  4 -4 1", "1  1/2  -2 1", "1  3/3  4 -4 1", 1);
    test_sub("3  0 1 1/3  1 2 1", "2  0 -1/2  1 1", "2  0 2/2  1 1", 1);
    test_sub("2  1 1/2  0 1", "2  -1 1/2  0 1", "1  2/2  0 1", 1);
    test_sub("1  1/3  3 5 2", "1  1/3  6 7 2", "1  1/4  6 13 9 2", 1);
    test_sub("2  1 1/2  0 2", "2  1 1/2  0 2", "0", 1);
    test_sub("2  -1 2/2  0 1", "2  -1 1/2  0 1", "1  1", 1);
    
    test_sub_in_place1("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "5  1 0 3 0 1/4  0 1 0 1", 1);
    test_sub_in_place1("3  -1 0 1/2  1 1", "1  2/2  -1 1", "3  -1 -2 1/2  -1 1", 1);
    test_sub_in_place1("0/2  1 1", "1  2/1  1", "1  -2", 1);
    test_sub_in_place1("1  -3/1  4", "0/3  1 0 1", "1  -3/1  4", 1);
    test_sub_in_place1("2  1 1/1  1", "2  -1 1/1  1", "1  2", 1);
    test_sub_in_place1("2  1 1/2  0 1", "2  2 1/2  -1 1", "2  -1 -2/3  0 -1 1", 1);
    test_sub_in_place1("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "4  -9 -12 -5 -1/3  2 3 1", 1);
    
    test_sub_in_place2("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "5  1 0 3 0 1/4  0 1 0 1", 1);
    test_sub_in_place2("3  -1 0 1/2  1 1", "1  2/2  -1 1", "3  -1 -2 1/2  -1 1", 1);
    test_sub_in_place2("0/2  1 1", "1  2/1  1", "1  -2", 1);
    test_sub_in_place2("1  -3/1  4", "0/3  1 0 1", "1  -3/1  4", 1);
    test_sub_in_place2("2  1 1/1  1", "2  -1 1/1  1", "1  2", 1);
    test_sub_in_place2("2  1 1/2  0 1", "2  2 1/2  -1 1", "2  -1 -2/3  0 -1 1", 1);
    test_sub_in_place2("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "4  -9 -12 -5 -1/3  2 3 1", 1);
    
    test_sub_in_place3("2  -1 1/2  2 1", "0", 1);
    
    // addmul
    test_addmul("1  1/2  0 2", "2  3 1/1  4", "3  1 0 1/4  -2 0 0 1", "5  -4 3 1 5 1/5  0 -8 0 0 4", 1);
    
    // submul
    test_submul("1  1/2  0 2", "2  3 1/1  4", "3  1 0 1/4  -2 0 0 1", "5  -4 -3 -1 -1 -1/5  0 -8 0 0 4", 1);
    
    //-----------------------------------------------------------------------//
    // Scalar multiplication and devision                                    //
    //-----------------------------------------------------------------------//
    
    mpz_init_set_si(mpzzero, 0);
    mpz_init_set_si(mpzone, 1);
    mpz_init_set_si(mpztwo, 2);
    
    mpq_init(mpqzero); mpq_set_si(mpqzero, 0, 1);
    mpq_init(mpqone); mpq_set_si(mpqone, 1, 1);
    mpq_init(mpqtwo); mpq_set_si(mpqtwo, 2, 1);
    mpq_init(mpqtwoinv); mpq_set_si(mpqtwoinv, 1, 2);
    
    // scalar_mul_si
    test_scalar_mul_si("0", 1, "0", 1);
    test_scalar_mul_si("0", 0, "0", 1);
    test_scalar_mul_si("1  2", 0, "0", 1);
    test_scalar_mul_si("1  1/1  2", -2, "1  -1", 1);
    test_scalar_mul_si("2  1 1/2  -2 3", 5, "2  5 5/2  -2 3", 1);
    test_scalar_mul_si("2  1 1/2  -2 2", 3, "2  3 3/2  -2 2", 1);
    
    // scalar_mul_mpz
    test_scalar_mul_mpz("0", mpzone, "0", 1);
    test_scalar_mul_mpz("0", mpzzero, "0", 1);
    test_scalar_mul_mpz("1  2", mpzzero, "0", 1);
    test_scalar_mul_mpz("1  1/1  2", mpztwo, "1  1", 1);
    
    // scalar_mul_mpq
    test_scalar_mul_mpq("0", mpqone, "0", 1);
    test_scalar_mul_mpq("0", mpqzero, "0", 1);
    test_scalar_mul_mpq("1  2", mpqzero, "0", 1);
    test_scalar_mul_mpq("1  1/1  2", mpqtwo, "1  1", 1);
    test_scalar_mul_mpq("1  -2/1  1", mpqtwoinv, "1  -1", 1);
    
    // scalar_div_si
    test_scalar_div_si("0", 1, "0", 1);
    test_scalar_div_si("1  2", 2, "1  1", 1);
    test_scalar_div_si("1  1/1  2", -2, "1  -1/1  4", 1);
    test_scalar_div_si("3  -5 0 3/2  1 1", 2, "3  -5 0 3/2  2 2", 1);
    test_scalar_div_si("3  2 8 4/2  0 1", 3, "3  2 8 4/2  0 3", 1);
    test_scalar_div_si("3  2 8 4/2  0 1", -3, "3  -2 -8 -4/2  0 3", 1);
    test_scalar_div_si("3  -27 0 9/2  0 1", -3, "3  9 0 -3/2  0 1", 1);
    
    // scalar_div_mpz
    test_scalar_div_mpz("0", mpzone, "0", 1);
    test_scalar_div_mpz("1  2", mpztwo, "1  1", 1);
    test_scalar_div_mpz("1  1/1  2", mpztwo, "1  1/1  4", 1);
    
    // scalar_div_mpq
    test_scalar_div_mpq("0", mpqone, "0", 1);
    test_scalar_div_mpq("1  2", mpqone, "1  2", 1);
    test_scalar_div_mpq("1  1/1  2", mpqtwo, "1  1/1  4", 1);
    test_scalar_div_mpq("1  -2/1  1", mpqtwoinv, "1  -4", 1);
    
    mpz_clear(mpzzero);
    mpz_clear(mpzone);
    mpz_clear(mpztwo);
    mpq_clear(mpqzero);
    mpq_clear(mpqone);
    mpq_clear(mpqtwo);
    mpq_clear(mpqtwoinv);
    
    //-----------------------------------------------------------------------//
    // Multiplication, division and powering                                 //
    //-----------------------------------------------------------------------//
    
    // mul
    test_mul("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "1  -1", 1);
    test_mul("3  -1 0 1/2  1 1", "1  2/2  -1 1", "1  2", 1);
    test_mul("0/2  1 1", "1  2/1  1", "0", 1);
    test_mul("1  -3/1  4", "0/3  1 0 1", "0", 1);
    test_mul("2  1 1/1  1", "2  -1 1/1  1", "3  -1 0 1", 1);
    test_mul("2  1 1/2  0 1", "2  2 1/2  -1 1", "3  2 3 1/3  0 -1 1", 1);
    test_mul("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "3  -2 1 1/2  1 1", 1);
    
    test_mul_in_place1("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "1  -1", 1);
    test_mul_in_place1("3  -1 0 1/2  1 1", "1  2/2  -1 1", "1  2", 1);
    test_mul_in_place1("0/2  1 1", "1  2/1  1", "0", 1);
    test_mul_in_place1("1  -3/1  4", "0/3  1 0 1", "0", 1);
    test_mul_in_place1("2  1 1/1  1", "2  -1 1/1  1", "3  -1 0 1", 1);
    test_mul_in_place1("2  1 1/2  0 1", "2  2 1/2  -1 1", "3  2 3 1/3  0 -1 1", 1);
    test_mul_in_place1("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "3  -2 1 1/2  1 1", 1);
    
    test_mul_in_place2("3  1 0 1/2  0 1", "2  0 -1/3  1 0 1", "1  -1", 1);
    test_mul_in_place2("3  -1 0 1/2  1 1", "1  2/2  -1 1", "1  2", 1);
    test_mul_in_place2("0/2  1 1", "1  2/1  1", "0", 1);
    test_mul_in_place2("1  -3/1  4", "0/3  1 0 1", "0", 1);
    test_mul_in_place2("2  1 1/1  1", "2  -1 1/1  1", "3  -1 0 1", 1);
    test_mul_in_place2("2  1 1/2  0 1", "2  2 1/2  -1 1", "3  2 3 1/3  0 -1 1", 1);
    test_mul_in_place2("2  -1 1/2  2 1", "3  4 4 1/2  1 1", "3  -2 1 1/2  1 1", 1);
    
    test_mul_in_place3("2  0 1/2  1 1", "3  0 0 1/3  1 2 1", 1);
    
    // div
    test_div("3  -1 0 1/1  2", "2  1 1/1  1", "2  -1 1/1  2", 1);
    test_div("0/2  1 1", "2  1 1/1  1", "0", 1);
    test_div("3  -1 0 1/1  4", "2  -1 -1/1  2", "2  1 -1/1  2", 1);
    test_div("2  1 1", "2  1 -1/2  1 -1", "2  1 1", 1);
    test_div("2  1 1/3  4 4 1", "2  -1 1/3  6 5 1", "3  3 4 1/3  -2 1 1", 1);
    
    test_div_in_place1("3  -1 0 1/1  2", "2  1 1/1  1", "2  -1 1/1  2", 1);
    test_div_in_place1("0/2  1 1", "2  1 1/1  1", "0", 1);
    test_div_in_place1("3  -1 0 1/1  4", "2  -1 -1/1  2", "2  1 -1/1  2", 1);
    test_div_in_place1("2  1 1", "2  1 -1/2  1 -1", "2  1 1", 1);
    test_div_in_place1("2  1 1/3  4 4 1", "2  -1 1/3  6 5 1", "3  3 4 1/3  -2 1 1", 1);
    test_div_in_place1("0", "1  2/2  3 5", "0", 1);
    
    test_div_in_place2("3  -1 0 1/1  2", "2  1 1/1  1", "2  -1 1/1  2", 1);
    test_div_in_place2("0/2  1 1", "2  1 1/1  1", "0", 1);
    test_div_in_place2("3  -1 0 1/1  4", "2  -1 -1/1  2", "2  1 -1/1  2", 1);
    test_div_in_place2("2  1 1", "2  1 -1/2  1 -1", "2  1 1", 1);
    test_div_in_place2("2  1 1/3  4 4 1", "2  -1 1/3  6 5 1", "3  3 4 1/3  -2 1 1", 1);
    
    test_div_in_place3("3  -1 0 1/1  2", "1  1", 1);
    
    // power
    test_power("2  0 -1/1  2", 3, "4  0 0 0 -1/1  8", 1);
    test_power("0", 0, "1  1", 1);
    test_power("2  1 -1", 0, "1  1", 1);
    test_power("2  1 1/2  0 1", 0, "1  1", 1);
    
    //-----------------------------------------------------------------------//
    // Derivative                                                            //
    //-----------------------------------------------------------------------//
    
    // derivative
    test_derivative("0", "0", 1);
    test_derivative("1  2", "0", 1);
    test_derivative("1  -1/1  2", "0", 1);
    test_derivative("2  0 1", "1  1", 1);
    test_derivative("3  1 0 1", "2  0 2", 1);
    test_derivative("1  1/2  0 1", "1  -1/3  0 0 1", 1);
    test_derivative("2  2 1/2  -1 1", "1  -3/3  1 -2 1", 1);
    
    test_derivative("2  0 1/3  1 2 1", "2  1 -1/4  1 3 3 1", 1);

    /* Bug which allowed constant factors */
    test_derivative("3  5 1 -2/2  10 2", "3  0 -10 -1/3  25 10 1", 1);
    
    //-----------------------------------------------------------------------//
    // Evaluation                                                            //
    //-----------------------------------------------------------------------//
    
    // evaluate
    test_evaluate("1  1/1  2", -2, 3, "1/2", 1);
    test_evaluate("3  1 0 1/2  0 1", -1, 2, "-5/2", 1);
    test_evaluate("2  3 1/2  -1 1", 1, 1, "P", 1);
    test_evaluate("2  3 1/2  -1 1", 2, 3, "-11", 1);
    test_evaluate("2  3 1/2  -1 2", 1, 2, "P", 1);
    test_evaluate("2  1 1/2  -1 1", 2, 1, "3", 1);
    
    //-----------------------------------------------------------------------//
    // String methods                                                        //
    //-----------------------------------------------------------------------//
    
    // from_string  -- mostly tested indirectly
    fmpz_poly_q_init(qpoly1);
    ans = fmpz_poly_q_from_string(qpoly1, "1  3/xyz");
    if (ans != 0 || !fmpz_poly_q_is_zero(qpoly1))
    {
        printf("test_from_string: failed\n");
    }
    else
    {
        printf("test_from_string: passed\n");
    }
    fmpz_poly_q_clear(qpoly1);
    
    fmpz_poly_q_init(qpoly1);
    ans = fmpz_poly_q_from_string(qpoly1, "abc/1  3");
    if (ans != 0 || !fmpz_poly_q_is_zero(qpoly1))
    {
        printf("test_from_string: failed\n");
    }
    else
    {
        printf("test_from_string: passed\n");
    }
    fmpz_poly_q_clear(qpoly1);
    
    fmpz_poly_q_init(qpoly1);
    ans = fmpz_poly_q_from_string(qpoly1, "abc/xyz");
    if (ans != 0 || !fmpz_poly_q_is_zero(qpoly1))
    {
        printf("test_from_string: failed\n");
    }
    else
    {
        printf("test_from_string: passed\n");
    }
    fmpz_poly_q_clear(qpoly1);
    
    // to_string_pretty
    test_to_string_pretty("1  -3", "-3", 1);
    test_to_string_pretty("3  1 2 1", "t^2+2*t+1", 1);
    test_to_string_pretty("1  -2/2  1 1", "-2/(t+1)", 1);
    test_to_string_pretty("2  1 1/2  -1 1", "(t+1)/(t-1)", 1);
    test_to_string_pretty("2  1 1/1  2", "(t+1)/2", 1);
    test_to_string_pretty("1  1/1  2", "1/2", 1);
}

