/******************************************************************************

    Copyright (C) 2009, 2010 Sebastian Pancratz

    Distributed under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    http://www.gnu.org/licenses/

******************************************************************************/

#include "mpz_poly.h"

#include "fmpz_poly_q.h"

/**
 * \file     fmpz_poly_q.c
 * \brief    Fast implementation of the rational function field
 * \author   Sebastian Pancratz
 * \date     Oct 2009 -- Dec 2010
 * \version  1.4.0
 * 
 * \mainpage 
 * 
 * <center>
 * A FLINT-based implementation of the rational function field.
 * </center>
 * 
 * \section Overview
 * 
 * The module <tt>fmpz_poly_q</tt> provides functions for performing 
 * arithmetic on rational functions in \f$\mathbf{Q}(t)\f$, represented as 
 * quotients of integer polynomials of type <tt>fmpz_poly_t</tt>.  These 
 * functions start with the prefix <tt>fmpz_poly_q_</tt>.
 * 
 * Rational functions are stored in objects of type #fmpz_poly_q_t, 
 * which is an array of #fmpz_poly_q_struct's of length one.  This 
 * permits passing parameters of type #fmpz_poly_q_t by reference.  
 * We also define the type #fmpz_poly_q_ptr to be a pointer to 
 * #fmpz_poly_q_struct's.
 * 
 * The representation of a rational function as the quotient of two integer 
 * polynomials can be made canonical by demanding the numerator and 
 * denominator to be coprime (as integer polynomials) and the denominator to 
 * have positive leading coefficient.  As the only special case, we represent 
 * the zero function as \f$0/1\f$.  All arithmetic functions assume that the 
 * operands are in this canonical form, and canonicalize their result.  If the 
 * numerator or denominator is modified individually, for example using the 
 * methods in the group \ref NumDen, it is the user's responsibility to 
 * canonicalize the rational function using the function 
 * #fmpz_poly_q_canonicalize() if necessary.
 * 
 * All methods support aliasing of their inputs and outputs \e unless 
 * explicitly stated otherwise, subject to the following caveat.  If 
 * different rational functions (as objects in memory, not necessarily in the 
 * mathematical sense) share some of the underlying integer polynomial 
 * objects, the behaviour is undefined.
 * 
 * The basic arithmetic operations, addition, subtraction and multiplication, 
 * are all implemented using adapted versions of Henrici's algorithms, see
 * [H56].  Differentiation is implemented in a way slightly improving on the 
 * algorithm described in [H80].
 * 
 * \section Example  Simple example
 * 
 * The following example computes the product of two rational functions and 
 * prints the result.
 * 
 * \code
 * #include "fmpz_poly_q.h"
 * ...
 * char *str, *strf, *strg;
 * fmpz_poly_q_t f, g;
 * fmpz_poly_q_init(f);
 * fmpz_poly_q_init(g);
 * fmpz_poly_q_from_string(f, "2  1 3/1  2");
 * fmpz_poly_q_from_string(g, "1  3/2  2 7");
 * strf = fmpz_poly_q_to_string_pretty(f, "t");
 * strg = fmpz_poly_q_to_string_pretty(g, "t");
 * fmpz_poly_q_mul(f, f, g);
 * str  = fmpz_poly_q_to_string_pretty(f, "t");
 * printf("%s * %s = %s\n", strf, strg, str);
 * free(str);
 * free(strf);
 * free(strg);
 * fmpz_poly_q_clear(f);
 * fmpz_poly_q_clear(g);
 * \endcode
 * 
 * The output is:
 * \code
 * (3*t+1)/2 * 3/(7*t+2) = (9*t+3)/(14*t+4)
 * \endcode
 * 
 * \section Changelog  Version history
 * 
 * - 1.4.0
 *   - Include #fmpz_poly_q_print() and #fmpz_poly_q_print_pretty()
 *   - Fix a bug in #fmpz_poly_q_derivative() which leaves the derivative 
 *     in a non-canonical form
 *   - Fix a bug in #fmpz_poly_q_evaluate() which under-allocated a 
 *     <tt>fmpz_t</tt>
 *   - Add various <tt>_randtest</tt> functions, which provide random 
 *     (non-zero) rational functions, polynomials, and integers
 *   - Add lots of test code
 * - 1.3.0
 *   - Rename everything from <tt>qfmpz_poly_</tt> to <tt>fmpz_poly_q_</tt>
 *     in preparation for the integration into FLINT
 *   - Small change in the grouping of methods, to fit the better with 
 *     other modules in FLINT or GMP
 * - 1.2.2
 *   - Rename <tt>_fmpz_evaluate_horner_mpq</tt> to 
 *     <tt>_fmpz_poly_evaluate_mpq_horner</tt>, and remove the condition that 
 *     the length of the polynomial must fit into a <tt>long</tt> rather than 
 *     merely an <tt>unsigned long</tt>
 *   - Replacing <tt>exit</tt> statements with <tt>abort</tt> statements, 
 *     since Sage apparently handles these better
 * - 1.2.1
 *   - Remove dependency on <tt>assert.h</tt>, replacing assertions by 
 *     explicit checks followed by an error message and an <tt>exit</tt> 
 *     statement
 *   - Small change to <tt>_fmpz_poly_q_sub_in_place()</tt>
 *   - Minor documentation changes
 * - 1.2.0
 *   - Henrici-like algorithm for <tt>qfmpz_poly_derivative()</tt>
 *   - Changed from <tt>fmpz_poly_t</tt> to <tt>fmpz_poly_p</tt> for the 
 *     numerator and denominator
 *   - Direct access to the numerator and denominator
 *   - Increased code coverage through <tt>qfmpz_poly-test</tt> 
 *     (<tt>qfmpz_poly.h</tt> 100%, <tt>qfmpz_poly.c</tt> 94.78%)
 * - 1.1.0
 *   - Minor bugfix in <tt>qfmpz_poly_evaluate()</tt>
 * - 1.0.0
 *   - First complete version
 * 
 * \bug  The method #fmpz_poly_q_to_string_pretty() leaks memory because the 
 *       underlying FLINT 1.5.0 method does so.
 * 
 * \todo  Implement some form of lazy evaluation.  One strategy would be to 
 *        change the representation of a rational function in the case of the 
 *        zero function to <tt>NULL</tt>.  This gives savings on 
 *        initialization and checking whether a function is zero.  Note that 
 *        it is typically unlikely that an arithmetic operation on two 
 *        functions yields zero, which is the only case in which this design 
 *        choice yields extra work.
 *        Another approach would be to implement a delayed initialization of 
 *        both the numerator and denominator.
 *        In either case, the methods providing direct access to the 
 *        underlying polynomial objects would require more care.
 * 
 * \see [H56] - Peter Henrici, 
 *              <em>A Subroutine for Computations with Rational Numbers</em>, 
 *              J. ACM, \b 3, \b 1, 1956, 6-9, 
 *              http://doi.acm.org/10.1145/320815.320818
 * 
 * \see [H80] - Ellis Horowitz, 
 *              <em>Algorithms for Rational Function Arithmetic 
 *              Operations</em>, 
 *              Annual ACM Symposium on Theory of Computing:  Proceedings of 
 *              the Fourth Annual ACM Symposium on Theory of Computing 
 *              (Denver), 108-118, 1972,
 *              http://doi.acm.org/10.1145/800152.804903
 */

/**
 * \defgroup Definitions        Type definitions
 * \defgroup MemoryManagement   Memory management
 * \defgroup Randomisation      Randomisation
 * \defgroup NumDen             Accessing numerator and denominator
 * \defgroup Assignment         Assignment and basic manipulation
 * \defgroup Comparison         Comparison
 * \defgroup Addition           Addition and subtraction
 * \defgroup ScalarMul          Scalar multiplication and division
 * \defgroup Multiplication     Multiplication and division
 * \defgroup Powering           Powering
 * \defgroup Derivative         Derivative
 * \defgroup Evaluation         Evaluation
 * \defgroup StringConversions  String conversions and I/O
 */

///////////////////////////////////////////////////////////////////////////////
// Implementation                                                            //
///////////////////////////////////////////////////////////////////////////////

/**
 * \ingroup  Definitions
 * 
 * Transforms the element \c rop into canonical form.
 * 
 * This means that the numerator and denominator are coprime (as integer 
 * polynomials), and that the denominator has positive leading coefficient.
 *
 * Assumes that the denominator of \c rop is non-zero.
 */
void fmpz_poly_q_canonicalize(fmpz_poly_q_ptr rop)
{
    fmpz_poly_t gcd;
    
    // Assertion!
    if (fmpz_poly_length(rop->den) < 1)
    {
        printf("ERROR (fmpz_poly_q_canonicalize).  Denominator is zero.\n");
        abort();
    }
    
    if (_fmpz_poly_is_one(rop->den))
        return;
    
    fmpz_poly_init(gcd);
    fmpz_poly_gcd(gcd, rop->num, rop->den);
    if (!_fmpz_poly_is_unit(gcd))
    {
        fmpz_poly_div(rop->num, rop->num, gcd);
        fmpz_poly_div(rop->den, rop->den, gcd);
    }
    fmpz_poly_clear(gcd);
    
    if (fmpz_sgn(fmpz_poly_lead(rop->den)) < 0)
    {
        fmpz_poly_neg(rop->num, rop->num);
        fmpz_poly_neg(rop->den, rop->den);
    }
}

/**
 * \ingroup  Randomisation
 */
void fmpz_randtest(fmpz_t * f, fmpz_poly_q_randstate_t R, ulong bits)
{
    mpz_t g;
    mpz_init(g);
    mpz_rrandomb(g, R, bits);
    if (gmp_urandomb_ui(R, 1))
        mpz_clrbit(g, bits - 1);
    if (gmp_urandomb_ui(R, 1))
        mpz_neg(g, g);
    *f = fmpz_realloc(*f, mpz_size(g));
    mpz_to_fmpz(*f, g);
    mpz_clear(g);
}

/**
 * \ingroup  Randomisation
 */
void fmpz_randtest_not_zero(fmpz_t * f, fmpz_poly_q_randstate_t R, ulong bits)
{
    mpz_t g;
    mpz_init(g);
    mpz_rrandomb(g, R, bits);
    if (gmp_urandomb_ui(R, 1))
        mpz_clrbit(g, bits - 1);
    if (gmp_urandomb_ui(R, 1))
        mpz_neg(g, g);
    
    if (mpz_sgn(g) == 0)
        fmpz_set_si(*f, -123);
    else
    {
        *f = fmpz_realloc(*f, mpz_size(g));
        mpz_to_fmpz(*f, g);
    }
    
    mpz_clear(g);
}

/**
 * \ingroup  Randomisation
 */
void fmpz_poly_randtest(fmpz_poly_t poly, fmpz_poly_q_randstate_t R, 
                        ulong len, ulong bits)
{
    long i;
    mpz_t temp;
    mpz_poly_t mpoly;

    if (len == 0 || bits == 0)
    {
        fmpz_poly_zero(poly);
        return;
    }

    mpz_init(temp);
    mpz_poly_init(mpoly);
    mpz_poly_zero(mpoly);

    fmpz_poly_fit_length(poly, len);
   
    for (i = 0; i < len; i++)
    {
        mpz_rrandomb(temp, R, bits);
        if (gmp_urandomb_ui(R, 1))
            mpz_clrbit(temp, bits - 1);
        if (gmp_urandomb_ui(R, 1))
            mpz_neg(temp,temp);
        mpz_poly_set_coeff(mpoly, i, temp);
    }

    mpz_poly_to_fmpz_poly(poly, mpoly);

    mpz_clear(temp);
    mpz_poly_clear(mpoly);
}

/**
 * \ingroup  Randomisation
 */
void fmpz_poly_randtest_not_zero(fmpz_poly_t poly, fmpz_poly_q_randstate_t R, 
                                 ulong len, ulong bits)
{
    if (len == 0ul)
        len = 1ul;
    if (bits == 0ul)
        bits = 23ul;
    fmpz_poly_randtest(poly, R, len, bits);
    if (poly->length == 0ul)
        fmpz_poly_set_coeff_si(poly, 0, 1);
}

/**
 * \ingroup  Randomisation
 */
void
fmpz_poly_q_randtest(fmpz_poly_q_t poly, fmpz_poly_q_randstate_t R, 
                     ulong len1, ulong bits1, ulong len2, ulong bits2)
{
    fmpz_poly_randtest(poly->num, R, len1, bits1);
    fmpz_poly_randtest_not_zero(poly->den, R, len2, bits2);
    fmpz_poly_q_canonicalize(poly);
} 

/**
 * \ingroup  Randomisation
 */
void 
fmpz_poly_q_randtest_not_zero(fmpz_poly_q_t poly, fmpz_poly_q_randstate_t R, 
                              ulong len1, ulong bits1, ulong len2, ulong bits2)
{
    fmpz_poly_randtest_not_zero(poly->num, R, len1, bits1);
    fmpz_poly_randtest_not_zero(poly->den, R, len2, bits2);
    fmpz_poly_q_canonicalize(poly);
}

/**
 * \ingroup  Addition
 * 
 * Sets \c rop to the sum of \c rop and \c op.
 */
void _fmpz_poly_q_add_in_place(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op)
{
    fmpz_poly_t d, poly, r2, s2;
    
    // If rop == op, return 2 * op
    if (rop == op)
    {
        fmpz_poly_q_scalar_mul_si(rop, rop, 2l);
        return;
    }
    
    // Is one of the numerators zero?
    if (fmpz_poly_degree(rop->num) < 0)
    {
        fmpz_poly_q_set(rop, op);
        return;
    }
    if (fmpz_poly_degree(op->num) < 0)
    {
        return;
    }
    
    // Is one or are both denominators equal to one?
    if (_fmpz_poly_is_one(rop->den))
    {
        if (_fmpz_poly_is_one(op->den))
        {
            fmpz_poly_add(rop->num, rop->num, op->num);
            return;
        }
        else
        {
            fmpz_poly_mul(rop->num, rop->num, op->den);
            fmpz_poly_add(rop->num, rop->num, op->num);
            fmpz_poly_set(rop->den, op->den);
            return;
        }
    }
    if (_fmpz_poly_is_one(op->den))
    {
        fmpz_poly_init(poly);
        fmpz_poly_mul(poly, rop->den, op->num);
        fmpz_poly_add(rop->num, rop->num, poly);
        fmpz_poly_clear(poly);
        return;
    }
    
    // Henrici's algorithm for summation in quotient fields   
    fmpz_poly_init(d);
    fmpz_poly_gcd(d, rop->den, op->den);
    if (fmpz_sgn(fmpz_poly_lead(d)) < 0)
        fmpz_poly_neg(d, d);
    
    if (_fmpz_poly_is_one(d))
    {
        fmpz_poly_mul(rop->num, rop->num, op->den);
        fmpz_poly_mul(d, rop->den, op->num);  // Using d as temp
        fmpz_poly_add(rop->num, rop->num, d);
        fmpz_poly_mul(rop->den, rop->den, op->den);
    }
    else
    {
        fmpz_poly_init(r2);
        fmpz_poly_init(s2);
        
        fmpz_poly_div(r2, rop->den, d);
        fmpz_poly_div(s2, op->den, d);
        
        fmpz_poly_mul(rop->num, rop->num, s2);
        fmpz_poly_mul(s2, op->num, r2);  // Using s2 as temp
        fmpz_poly_add(rop->num, rop->num, s2);
        
        if (fmpz_poly_degree(rop->num) < 0)
        {
            fmpz_poly_zero(rop->den);
            fmpz_poly_set_coeff_si(rop->den, 0, 1);
        }
        else
        {
            fmpz_poly_mul(rop->den, r2, op->den);
            
            fmpz_poly_gcd(r2, rop->num, d);  // Using r2 as temp
            if (fmpz_sgn(fmpz_poly_lead(r2)) < 0)
                fmpz_poly_neg(r2, r2);
            
            if (!_fmpz_poly_is_one(r2))
            {
                fmpz_poly_div(rop->num, rop->num, r2);
                fmpz_poly_div(rop->den, rop->den, r2);
            }
        }
        fmpz_poly_clear(r2);
        fmpz_poly_clear(s2);
    }
    fmpz_poly_clear(d);
}

/**
 * \ingroup  Addition
 * 
 * Sets \c rop to the sum of \c op1 and \c op2.
 */
void fmpz_poly_q_add(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2)
{
    fmpz_poly_t d, r2, s2;
    
    // If op1 == op2, return 2 * op1
    if (op1 == op2)
    {
        fmpz_poly_q_scalar_mul_si(rop, op1, 2l);
        return;
    }
    
    // Is rop == op1 or rop == op2?
    if (rop == op1)
    {
        _fmpz_poly_q_add_in_place(rop, op2);
        return;
    }
    if (rop == op2)
    {
        _fmpz_poly_q_add_in_place(rop, op1);
        return;
    }
    
    // From here on, we may assume that rop, op1 and op2 all refer to 
    // distinct objects in memory, although they may still be equal
    
    // Is one of the numerators zero?
    if (fmpz_poly_degree(op1->num) < 0)
    {
        fmpz_poly_q_set(rop, op2);
        return;
    }
    if (fmpz_poly_degree(op2->num) < 0)
    {
        fmpz_poly_q_set(rop, op1);
        return;
    }
    
    // Is one or are even both denominators equal to one?
    if (_fmpz_poly_is_one(op1->den))
    {
        if (_fmpz_poly_is_one(op2->den))
        {
            fmpz_poly_add(rop->num, op1->num, op2->num);
            fmpz_poly_zero(rop->den);
            fmpz_poly_set_coeff_si(rop->den, 0, 1);
            return;
        }
        else
        {
            fmpz_poly_mul(rop->num, op1->num, op2->den);
            fmpz_poly_add(rop->num, rop->num, op2->num);
            fmpz_poly_set(rop->den, op2->den);
            return;
        }
    }
    if (_fmpz_poly_is_one(op2->den))
    {
        fmpz_poly_mul(rop->num, op2->num, op1->den);
        fmpz_poly_add(rop->num, op1->num, rop->num);
        fmpz_poly_set(rop->den, op1->den);
        return;
    }
    
    // Henrici's algorithm for summation in quotient fields
    
    // We begin by using rop->num as a temporary variable for the gcd of the 
    // two denominators' greatest common divisor
    fmpz_poly_gcd(rop->num, op1->den, op2->den);
    if (fmpz_sgn(fmpz_poly_lead(rop->num)) < 0)
        fmpz_poly_neg(rop->num, rop->num);
    
    if (_fmpz_poly_is_one(rop->num))
    {
        fmpz_poly_mul(rop->num, op1->num, op2->den);
        fmpz_poly_mul(rop->den, op1->den, op2->num);  // Using rop->den as temp
        fmpz_poly_add(rop->num, rop->num, rop->den);
        fmpz_poly_mul(rop->den, op1->den, op2->den);
    }
    else
    {
        // We now copy rop->num into a new variable d, so we no longer need 
        // rop->num as a temporary variable
        fmpz_poly_init(d);
        fmpz_poly_swap(d, rop->num);
        
        fmpz_poly_init(r2);
        fmpz_poly_init(s2);
        
        fmpz_poly_div(r2, op1->den, d);  // +ve leading coeff
        fmpz_poly_div(s2, op2->den, d);  // +ve leading coeff
        
        fmpz_poly_mul(rop->num, op1->num, s2);
        fmpz_poly_mul(rop->den, op2->num, r2);  // Using rop->den as temp
        fmpz_poly_add(rop->num, rop->num, rop->den);
        
        if (fmpz_poly_degree(rop->num) < 0)
        {
            fmpz_poly_zero(rop->den);
            fmpz_poly_set_coeff_si(rop->den, 0, 1);
        }
        else
        {
            fmpz_poly_mul(rop->den, op1->den, s2);
            
            fmpz_poly_gcd(r2, rop->num, d);
            if (fmpz_sgn(fmpz_poly_lead(r2)) < 0)
                fmpz_poly_neg(r2, r2);
            
            if (!_fmpz_poly_is_one(r2))
            {
                fmpz_poly_div(rop->num, rop->num, r2);
                fmpz_poly_div(rop->den, rop->den, r2);
            }
        }
        
        fmpz_poly_clear(d);
        fmpz_poly_clear(r2);
        fmpz_poly_clear(s2);
    }
}

/**
 * \ingroup  Addition
 * 
 * Sets \c rop to the difference of \c rop and \c op.
 */
void _fmpz_poly_q_sub_in_place(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op)
{
    if (rop == op)
    {
        fmpz_poly_q_zero(rop);
        return;
    }
    
    fmpz_poly_q_neg(rop, rop);
    _fmpz_poly_q_add_in_place(rop, op);
    fmpz_poly_q_neg(rop, rop);
}

/**
 * \ingroup  Addition
 * 
 * Sets \c rop to the difference of \c op1 and \c op2.
 */
void fmpz_poly_q_sub(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2)
{
    fmpz_poly_t d, r2, s2;
    
    if (op1 == op2)
    {
        fmpz_poly_q_zero(rop);
        return;
    }
    if (rop == op1)
    {
        _fmpz_poly_q_sub_in_place(rop, op2);
        return;
    }
    if (rop == op2)
    {
        _fmpz_poly_q_sub_in_place(rop, op1);
        fmpz_poly_q_neg(rop, rop);
        return;
    }
   
    // From here on, we know that rop, op1 and op2 refer to distinct objects 
    // in memory, although as rational functions they may still be equal
    
    // XXX:  Do not maintain the remaining part of the function separately!!!
    //       Instead, note that this is very similar to the corresponding part 
    //       of the summation code.
    
    // Is one of the numerators zero?
    if (fmpz_poly_degree(op1->num) < 0)
    {
        fmpz_poly_q_neg(rop, op2);
        return;
    }
    if (fmpz_poly_degree(op2->num) < 0)
    {
        fmpz_poly_q_set(rop, op1);
        return;
    }
    
    // Is one or are even both denominators equal to one?
    if (_fmpz_poly_is_one(op1->den))
    {
        if (_fmpz_poly_is_one(op2->den))
        {
            fmpz_poly_sub(rop->num, op1->num, op2->num);
            fmpz_poly_zero(rop->den);
            fmpz_poly_set_coeff_si(rop->den, 0, 1);
            return;
        }
        else
        {
            fmpz_poly_mul(rop->num, op1->num, op2->den);
            fmpz_poly_sub(rop->num, rop->num, op2->num);
            fmpz_poly_set(rop->den, op2->den);
            return;
        }
    }
    if (_fmpz_poly_is_one(op2->den))
    {
        fmpz_poly_mul(rop->num, op2->num, op1->den);
        fmpz_poly_sub(rop->num, op1->num, rop->num);
        fmpz_poly_set(rop->den, op1->den);
        return;
    }
    
    // Henrici's algorithm for summation in quotient fields
    
    // We begin by using rop->num as a temporary variable for the gcd of the 
    // two denominators' greatest common divisor
    fmpz_poly_gcd(rop->num, op1->den, op2->den);
    if (fmpz_sgn(fmpz_poly_lead(rop->num)) < 0)
        fmpz_poly_neg(rop->num, rop->num);
    
    if (_fmpz_poly_is_one(rop->num))
    {
        fmpz_poly_mul(rop->num, op1->num, op2->den);
        fmpz_poly_mul(rop->den, op1->den, op2->num);  // Using rop->den as temp
        fmpz_poly_sub(rop->num, rop->num, rop->den);
        fmpz_poly_mul(rop->den, op1->den, op2->den);
    }
    else
    {
        // We now copy rop->num into a new variable d, so we no longer need 
        // rop->num as a temporary variable
        fmpz_poly_init(d);
        fmpz_poly_swap(d, rop->num);
        
        fmpz_poly_init(r2);
        fmpz_poly_init(s2);
        
        fmpz_poly_div(r2, op1->den, d);  // +ve leading coeff
        fmpz_poly_div(s2, op2->den, d);  // +ve leading coeff
        
        fmpz_poly_mul(rop->num, op1->num, s2);
        fmpz_poly_mul(rop->den, op2->num, r2);  // Using rop->den as temp
        fmpz_poly_sub(rop->num, rop->num, rop->den);
        
        if (fmpz_poly_degree(rop->num) < 0)
        {
            fmpz_poly_zero(rop->den);
            fmpz_poly_set_coeff_si(rop->den, 0, 1);
        }
        else
        {
            fmpz_poly_mul(rop->den, op1->den, s2);
            
            fmpz_poly_gcd(r2, rop->num, d);
            if (fmpz_sgn(fmpz_poly_lead(r2)) < 0)
                fmpz_poly_neg(r2, r2);
            
            if (!_fmpz_poly_is_one(r2))
            {
                fmpz_poly_div(rop->num, rop->num, r2);
                fmpz_poly_div(rop->den, rop->den, r2);
            }
        }
        
        fmpz_poly_clear(d);
        fmpz_poly_clear(r2);
        fmpz_poly_clear(s2);
    }
}

/**
 * \ingroup  Addition
 * 
 * Sets \c rop to <tt>rop + op1 * op2</tt>.
 * 
 * Currently, this method refers to the methods #fmpz_poly_q_mul() and 
 * #fmpz_poly_q_add() to form the result in the naive way.
 * 
 * \todo  Implement this method more efficiently.
 */
void fmpz_poly_q_addmul(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2)
{
    fmpz_poly_q_t temp;
    fmpz_poly_q_init(temp);
    fmpz_poly_q_mul(temp, op1, op2);
    fmpz_poly_q_add(rop, rop, temp);
    fmpz_poly_q_clear(temp);
}

/**
 * \ingroup  Addition
 * 
 * Sets \c rop to <tt>rop - op1 * op2</tt>.
 * 
 * Currently, this method refers to the methods #fmpz_poly_q_mul() and 
 * #fmpz_poly_q_sub() to form the result in the naive way.
 * 
 * \todo  Implement this method more efficiently.
 */
void fmpz_poly_q_submul(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2)
{
    fmpz_poly_q_t temp;
    fmpz_poly_q_init(temp);
    fmpz_poly_q_mul(temp, op1, op2);
    fmpz_poly_q_sub(rop, rop, temp);
    fmpz_poly_q_clear(temp);
}

/**
 * \ingroup  ScalarMul
 * 
 * Sets \c rop to the scalar product of \c op and the integer \c x.
 */
void fmpz_poly_q_scalar_mul_si(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, long x)
{
    fmpz_t cont, fx, gcd;
    
    if (x == 1l)
    {
        fmpz_poly_q_set(rop, op);
        return;
    }
    
    cont = fmpz_init(op->den->limbs);
    fmpz_poly_content(cont, op->den);
    if (fmpz_sgn(cont) < 0)
        fmpz_neg(cont, cont);
    
    if (fmpz_is_one(cont))
    {
        fmpz_poly_scalar_mul_si(rop->num, op->num, x);
        fmpz_poly_set(rop->den, op->den);
        fmpz_clear(cont);
        return;
    }
    
    fx = fmpz_init(1);
    fmpz_set_si(fx, x);
    
    gcd = fmpz_init(FLINT_MAX(op->den->limbs, fmpz_size(fx)));
    fmpz_gcd(gcd, cont, fx);
    if (fmpz_sgn(gcd) < 0)
        fmpz_neg(gcd, gcd);
    
    if (fmpz_is_one(gcd))
    {
        fmpz_poly_scalar_mul_si(rop->num, op->num, x);
        fmpz_poly_set(rop->den, op->den);
    }
    else
    {
        fmpz_tdiv(fx, fx, gcd);
        fmpz_poly_scalar_mul_fmpz(rop->num, op->num, fx);
        fmpz_poly_scalar_div_fmpz(rop->den, op->den, gcd);
    }
    
    fmpz_clear(cont);
    fmpz_clear(fx);
    fmpz_clear(gcd);
}

/**
 * \ingroup  ScalarMul
 * 
 * Sets \c rop to the scalar multiple of \c op with the \c mpz_t \c x.
 * 
 * \todo  The implementation of this method could be improved, rewriting it in 
 *        a similar way to #fmpz_poly_q_scalar_mul_si().
 */
void fmpz_poly_q_scalar_mul_mpz(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, const mpz_t x)
{
    fmpz_poly_scalar_mul_mpz(rop->num, op->num, x);
    fmpz_poly_set(rop->den, op->den);
    fmpz_poly_q_canonicalize(rop);
}

/**
 * \ingroup  ScalarMul
 * 
 * Sets \c rop to the scalar multiple of \c op with the \c mpq_t \c x.
 * 
 * \todo  The implementation of this method could be improved, rewriting it in 
 *        a similar way to #fmpz_poly_q_scalar_mul_si().
 */
void fmpz_poly_q_scalar_mul_mpq(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, const mpq_t x)
{
    fmpz_poly_scalar_mul_mpz(rop->num, op->num, mpq_numref(x));
    fmpz_poly_scalar_mul_mpz(rop->den, op->den, mpq_denref(x));
    fmpz_poly_q_canonicalize(rop);
}

/**
 * \ingroup  ScalarMul
 * 
 * Sets \c rop to the scalar multiple of \c op with the multiplicative inverse 
 * of the integer \c x.
 *
 * Assumes that \c x is non-zero.
 */
void fmpz_poly_q_scalar_div_si(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, long x)
{
    fmpz_t cont, fx, gcd;
    
    // Assertion!
    if (x == 0l)
    {
        printf("ERROR (fmpz_poly_q_scalar_div_si).  Division by zero.\n");
        abort();
    }
    
    if (x == 1l)
    {
        fmpz_poly_q_set(rop, op);
        return;
    }
    
    cont = fmpz_init(op->num->limbs);
    fmpz_poly_content(cont, op->num);
    if (fmpz_sgn(cont) < 0)
        fmpz_neg(cont, cont);
    
    if (fmpz_is_one(cont))
    {
        if (x > 0l)
        {
            fmpz_poly_set(rop->num, op->num);
            fmpz_poly_scalar_mul_si(rop->den, op->den, x);
        }
        else
        {
            fmpz_poly_neg(rop->num, op->num);
            fmpz_poly_scalar_mul_si(rop->den, op->den, -x);
        }
        fmpz_clear(cont);
        return;
    }
    
    fx = fmpz_init(1);
    fmpz_set_si(fx, x);
    
    gcd = fmpz_init(FLINT_MAX(op->num->limbs, fmpz_size(fx)));
    fmpz_gcd(gcd, cont, fx);
    if (fmpz_sgn(gcd) < 0)
        fmpz_neg(gcd, gcd);
    
    if (fmpz_is_one(gcd))
    {
        if (x > 0l)
        {
            fmpz_poly_set(rop->num, op->num);
            fmpz_poly_scalar_mul_si(rop->den, op->den, x);
        }
        else
        {
            fmpz_poly_neg(rop->num, op->num);
            fmpz_poly_scalar_mul_si(rop->den, op->den, -x);
        }
    }
    else
    {
        fmpz_poly_scalar_div_fmpz(rop->num, op->num, gcd);
        fmpz_tdiv(fx, fx, gcd);
        fmpz_poly_scalar_mul_fmpz(rop->den, op->den, fx);
        if (x < 0l)
        {
            fmpz_poly_neg(rop->num, rop->num);
            fmpz_poly_neg(rop->den, rop->den);
        }
    }
    
    fmpz_clear(cont);
    fmpz_clear(fx);
    fmpz_clear(gcd);
}

/**
 * \ingroup  ScalarMul
 * 
 * Sets \c rop to the scalar multiple of \c op with the multiplicative inverse 
 * of the integer \c x.
 *
 * Assumes that the integer \c x is non-zero.
 *
 * \todo  The implementation of this method could be improved, rewriting it in 
 *        a similar way to #fmpz_poly_q_scalar_div_si().
 */
void fmpz_poly_q_scalar_div_mpz(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, const mpz_t x)
{
    // Assertion!
    if (mpz_sgn(x) == 0)
    {
        printf("ERROR (fmpz_poly_q_scalar_div_mpz).  Division by zero.\n");
        abort();
    }
    
    fmpz_poly_set(rop->num, op->num);
    fmpz_poly_scalar_mul_mpz(rop->den, op->den, x);
    fmpz_poly_q_canonicalize(rop);
}

/**
 * \ingroup  ScalarMul
 * 
 * Sets \c rop to the scalar multiple of \c op with the multiplicative inverse 
 * of the rational \c x.
 *
 * Assumes that the rational \c x is in lowest terms and non-zero.
 *
 * \todo  The implementation of this method could be improved, rewriting it in 
 *        a similar way to #fmpz_poly_q_scalar_div_si().
 */
void fmpz_poly_q_scalar_div_mpq(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, const mpq_t x)
{
    // Assertion!
    if (mpz_sgn(mpq_numref(x)) == 0)
    {
        printf("ERROR (fmpz_poly_q_scalar_div_mpq).  Division by zero.\n");
        abort();
    }
    
    fmpz_poly_scalar_mul_mpz(rop->num, op->num, mpq_denref(x));
    fmpz_poly_scalar_mul_mpz(rop->den, op->den, mpq_numref(x));
    fmpz_poly_q_canonicalize(rop);
}

/**
 * \ingroup  Multiplication
 * 
 * Sets \c rop to the product of \c rop with \c op.
 */
void _fmpz_poly_q_mul_in_place(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op)
{
    fmpz_poly_t d, e;
    
    if (rop == op)
    {
        fmpz_poly_power(rop->num, op->num, 2ul);
        fmpz_poly_power(rop->den, op->den, 2ul);
        return;
    }
    
    if (fmpz_poly_degree(rop->num) < 0 || fmpz_poly_degree(op->num) < 0)
    {
        fmpz_poly_q_zero(rop);
        return;
    }
    
    // From here on, rop and op point to two different object in memory, and 
    // these are both non-zero rational functions
    
    fmpz_poly_init(d);
    fmpz_poly_gcd(d, rop->num, op->den);
    if (fmpz_sgn(fmpz_poly_lead(d)) < 0)
        fmpz_poly_neg(d, d);
    
    if (_fmpz_poly_is_one(d))
    {
        fmpz_poly_gcd(d, rop->den, op->num);
        if (fmpz_sgn(fmpz_poly_lead(d)) < 0)
            fmpz_poly_neg(d, d);
        
        if (_fmpz_poly_is_one(d))
        {
            fmpz_poly_mul(rop->num, rop->num, op->num);
            fmpz_poly_mul(rop->den, rop->den, op->den);
        }
        else
        {
            fmpz_poly_div(rop->den, rop->den, d);
            fmpz_poly_mul(rop->den, rop->den, op->den);
            fmpz_poly_div(d, op->num, d);
            fmpz_poly_mul(rop->num, rop->num, d);
        }
    }
    else
    {
        fmpz_poly_init(e);
        fmpz_poly_gcd(e, rop->den, op->num);
        if (fmpz_sgn(fmpz_poly_lead(e)) < 0)
            fmpz_poly_neg(e, e);
        
        if (_fmpz_poly_is_one(e))
        {
            fmpz_poly_div(rop->num, rop->num, d);
            fmpz_poly_mul(rop->num, rop->num, op->num);
            fmpz_poly_div(d, op->den, d);
            fmpz_poly_mul(rop->den, rop->den, d);
        }
        else
        {
            fmpz_poly_div(rop->num, rop->num, d);
            fmpz_poly_div(rop->den, rop->den, e);
            fmpz_poly_div(e, op->num, e);
            fmpz_poly_div(d, op->den, d);
            fmpz_poly_mul(rop->num, rop->num, e);
            fmpz_poly_mul(rop->den, rop->den, d);
        }
        fmpz_poly_clear(e);
    }
    fmpz_poly_clear(d);
}

/**
 * \ingroup  Multiplication
 * 
 * Sets \c rop to the product of \c op1 and \c op2.
 */
void fmpz_poly_q_mul(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2)
{
    fmpz_poly_t t, u;
    
    if (op1 == op2)
    {
        fmpz_poly_power(rop->num, op1->num, 2ul);
        fmpz_poly_power(rop->den, op1->den, 2ul);
        return;
    }
    
    if (rop == op1)
    {
        _fmpz_poly_q_mul_in_place(rop, op2);
        return;
    }
    if (rop == op2)
    {
        _fmpz_poly_q_mul_in_place(rop, op1);
        return;
    }
    
    // Is one of the operands zero?
    if (fmpz_poly_q_is_zero(op1) || fmpz_poly_q_is_zero(op2))
    {
        fmpz_poly_q_zero(rop);
        return;
    }
    
    // From here on, we may assume that rop, op1 and op2 refer to distinct 
    // objects in memory, and that op1 and op2 are non-zero
    
    // Are both denominators equal to one?
    if (_fmpz_poly_is_one(op1->den) && _fmpz_poly_is_one(op2->den))
    {
        fmpz_poly_mul(rop->num, op1->num, op2->num);
        fmpz_poly_zero(rop->den);
        fmpz_poly_set_coeff_si(rop->den, 0, 1);
        return;
    } 
    
    fmpz_poly_gcd(rop->num, op1->num, op2->den);
    if (fmpz_sgn(fmpz_poly_lead(rop->num)) < 0)
        fmpz_poly_neg(rop->num, rop->num);
    
    if (_fmpz_poly_is_one(rop->num))
    {
        fmpz_poly_gcd(rop->den, op2->num, op1->den);
        if (fmpz_sgn(fmpz_poly_lead(rop->den)) < 0)
            fmpz_poly_neg(rop->den, rop->den);
        
        if (_fmpz_poly_is_one(rop->den))
        {
            fmpz_poly_mul(rop->num, op1->num, op2->num);
            fmpz_poly_mul(rop->den, op1->den, op2->den);
        }
        else
        {
            fmpz_poly_div(rop->num, op2->num, rop->den);
            fmpz_poly_mul(rop->num, op1->num, rop->num);
            fmpz_poly_div(rop->den, op1->den, rop->den);
            fmpz_poly_mul(rop->den, rop->den, op2->den);
        }
    }
    else
    {
        fmpz_poly_gcd(rop->den, op2->num, op1->den);
        if (fmpz_sgn(fmpz_poly_lead(rop->den)) < 0)
            fmpz_poly_neg(rop->den, rop->den);
        
        if (_fmpz_poly_is_one(rop->den))
        {
            fmpz_poly_div(rop->den, op2->den, rop->num);
            fmpz_poly_mul(rop->den, op1->den, rop->den);
            fmpz_poly_div(rop->num, op1->num, rop->num);
            fmpz_poly_mul(rop->num, rop->num, op2->num);
        }
        else
        {
            fmpz_poly_init(t);
            fmpz_poly_init(u);
            fmpz_poly_div(t, op1->num, rop->num);
            fmpz_poly_div(u, op2->den, rop->num);
            fmpz_poly_div(rop->num, op2->num, rop->den);
            fmpz_poly_mul(rop->num, t, rop->num);
            fmpz_poly_div(rop->den, op1->den, rop->den);
            fmpz_poly_mul(rop->den, rop->den, u);
            fmpz_poly_clear(t);
            fmpz_poly_clear(u);
        }
    }
}

/**
 * \ingroup  Multiplication
 * 
 * Sets \c rop to the quotient of \c rop and \c op.
 *
 * Assumes that \c op is non-zero.
 */
void _fmpz_poly_q_div_in_place(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op)
{
    // Assertion!
    if (fmpz_poly_q_is_zero(op))
    {
        printf("ERROR (_fmpz_poly_q_div_in_place).  Division by zero.\n");
        abort();
    }
    
    if (fmpz_poly_q_is_zero(rop))
        return;
    
    fmpz_poly_q_inv(rop, rop);
    _fmpz_poly_q_mul_in_place(rop, op);
    fmpz_poly_q_inv(rop, rop);
}

/**
 * \ingroup  Multiplication
 * 
 * Sets \c rop to the quotient of \c op1 and \c op2.
 *
 * Assumes that \c op2 is non-zero.
 */
void fmpz_poly_q_div(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2)
{
    fmpz_poly_t t, u;
    
    // Assertion!
    if (fmpz_poly_q_is_zero(op2))
    {
        printf("ERROR (fmpz_poly_q_div).  Division by zero.\n");
        abort();
    }
    
    if (op1 == op2)
    {
        fmpz_poly_zero(rop->num);
        fmpz_poly_set_coeff_si(rop->num, 0, 1);
        fmpz_poly_zero(rop->den);
        fmpz_poly_set_coeff_si(rop->den, 0, 1);
        return;
    }
    
    if (fmpz_poly_q_is_zero(op1))
    {
        fmpz_poly_q_zero(rop);
        return;
    }
    
    if (rop == op1)
    {
        _fmpz_poly_q_div_in_place(rop, op2);
        return;
    }
    if (rop == op2)
    {
        _fmpz_poly_q_div_in_place(rop, op1);
        fmpz_poly_q_inv(rop, rop);
        return;
    }
    
    // From here on, we know that rop, op1 and op2 refer to distinct objects 
    // in memory, and that op1 and op2 are non-zero rational functions
    
    // XXX:  Do not maintain the remaining part of the function separately!!!
    //       Instead, note that this is the same as the corresponding part of 
    //       the multiplication code, with op2->num and op2->den swapped.
    //       
    //       The only caveat to this is that we cannot assume the leading 
    //       coefficient of op2->num to be positive, and thus check for this 
    //       in the very end.
    
    // Are both denominators equal to one?
    if (_fmpz_poly_is_one(op1->den) && _fmpz_poly_is_one(op2->num))
    {
        fmpz_poly_mul(rop->num, op1->num, op2->den);
        fmpz_poly_zero(rop->den);
        fmpz_poly_set_coeff_si(rop->den, 0, 1);
        return;
    } 
    
    fmpz_poly_gcd(rop->num, op1->num, op2->num);
    if (fmpz_sgn(fmpz_poly_lead(rop->num)) < 0)
        fmpz_poly_neg(rop->num, rop->num);
    
    if (_fmpz_poly_is_one(rop->num))
    {
        fmpz_poly_gcd(rop->den, op2->den, op1->den);
        if (fmpz_sgn(fmpz_poly_lead(rop->den)) < 0)
            fmpz_poly_neg(rop->den, rop->den);
        
        if (_fmpz_poly_is_one(rop->den))
        {
            fmpz_poly_mul(rop->num, op1->num, op2->den);
            fmpz_poly_mul(rop->den, op1->den, op2->num);
        }
        else
        {
            fmpz_poly_div(rop->num, op2->den, rop->den);
            fmpz_poly_mul(rop->num, op1->num, rop->num);
            fmpz_poly_div(rop->den, op1->den, rop->den);
            fmpz_poly_mul(rop->den, rop->den, op2->num);
        }
    }
    else
    {
        fmpz_poly_gcd(rop->den, op2->den, op1->den);
        if (fmpz_sgn(fmpz_poly_lead(rop->den)) < 0)
            fmpz_poly_neg(rop->den, rop->den);
        
        if (_fmpz_poly_is_one(rop->den))
        {
            fmpz_poly_div(rop->den, op2->num, rop->num);
            fmpz_poly_mul(rop->den, op1->den, rop->den);
            fmpz_poly_div(rop->num, op1->num, rop->num);
            fmpz_poly_mul(rop->num, rop->num, op2->den);
        }
        else
        {
            fmpz_poly_init(t);
            fmpz_poly_init(u);
            fmpz_poly_div(t, op1->num, rop->num);
            fmpz_poly_div(u, op2->num, rop->num);
            fmpz_poly_div(rop->num, op2->den, rop->den);
            fmpz_poly_mul(rop->num, t, rop->num);
            fmpz_poly_div(rop->den, op1->den, rop->den);
            fmpz_poly_mul(rop->den, rop->den, u);
            fmpz_poly_clear(t);
            fmpz_poly_clear(u);
        }
    }
    
    // XXX:  Check that the numerator has the appropriate sign.
    if (fmpz_sgn(fmpz_poly_lead(rop->den)) < 0)
    {
        fmpz_poly_neg(rop->num, rop->num);
        fmpz_poly_neg(rop->den, rop->den);
    }
}

/**
 * \ingroup  Powering
 * 
 * Sets \c rop to the <tt>exp</tt>th power of \c op.
 *
 * The corner case of <tt>exp == 0</tt> is handled by setting \c rop to the 
 * constant function \f$1\f$.  Note that this includes the case \f$0^0 = 1\f$.
 */
void fmpz_poly_q_power(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, unsigned long exp)
{
    if (exp == 0ul)
    {
        fmpz_poly_q_one(rop);
    }
    else
    {
        fmpz_poly_power(rop->num, op->num, exp);
        fmpz_poly_power(rop->den, op->den, exp);
    }
}

/**
 * \ingroup  Derivative
 * 
 * Sets \c rop to the derivative of \c op.
 * 
 * \todo  The second argument should be declared \c const, but as of 
 *        FLINT 1.5.0 this generates a compile-time warning.
 */
void fmpz_poly_q_derivative(fmpz_poly_q_ptr rop, fmpz_poly_q_ptr op)
{
    fmpz_poly_t d, lhs, rhs;
    
    if (fmpz_poly_q_is_zero(op))
    {
        fmpz_poly_q_zero(rop);
        return;
    }
    
    if (fmpz_poly_length(op->den) == 1)
    {
        fmpz_poly_derivative(rop->num, op->num);
        fmpz_poly_set(rop->den, op->den);
        fmpz_poly_q_canonicalize(rop);
        return;
    }
    
    fmpz_poly_init(d);
    fmpz_poly_init(rhs);
    
    fmpz_poly_derivative(rhs, op->den);
    fmpz_poly_gcd(d, rhs, op->den);
    if (fmpz_sgn(fmpz_poly_lead(d)) < 0)
        fmpz_poly_neg(d, d);
    if (!_fmpz_poly_is_one(d))
        fmpz_poly_div(rhs, rhs, d);
    fmpz_poly_mul(rhs, op->num, rhs);
    
    fmpz_poly_derivative(rop->num, op->num);
    if (_fmpz_poly_is_one(d))
    {
        fmpz_poly_mul(rop->num, rop->num, op->den);
        fmpz_poly_power(rop->den, op->den, 2ul);
    }
    else
    {
        fmpz_poly_init(lhs);
        fmpz_poly_div(lhs, op->den, d);
        fmpz_poly_mul(rop->num, rop->num, lhs);
        fmpz_poly_mul(rop->den, op->den, lhs);
        fmpz_poly_clear(lhs);
    }
    fmpz_poly_sub(rop->num, rop->num, rhs);
    
    /* Canonicalize:  there can be at most a constant factor */
    {
        fmpz_t a, b, c;

        a = fmpz_init(rop->num->limbs);
        b = fmpz_init(rop->den->limbs);
        c = fmpz_init(FLINT_MAX(rop->num->limbs, rop->den->limbs));

        fmpz_poly_content(a, rop->num);
        fmpz_poly_content(b, rop->den);
        fmpz_gcd(c, a, b);
        if (fmpz_sgn(c) < 0)
            fmpz_neg(c, c);

        if (!fmpz_is_one(c))
        {
            fmpz_poly_scalar_div_fmpz(rop->num, rop->num, c);
            fmpz_poly_scalar_div_fmpz(rop->den, rop->den, c);
        }

        fmpz_clear(a);
        fmpz_clear(b);
        fmpz_clear(c);
    }
    
    fmpz_poly_clear(d);
    fmpz_poly_clear(rhs);
}

/**
 * \ingroup  Evaluation
 * 
 * Evaluates the integer polynomial \c f at the rational \c a using Horner's 
 * method.
 */
void _fmpz_poly_evaluate_mpq_horner(mpq_t rop, const fmpz_poly_t f, const mpq_t a)
{
    mpq_t temp;
    unsigned long n;
    
    n = fmpz_poly_length(f);
    
    if (n == 0)
    {
        mpq_set_si(rop, 0, 1);
    }
    else if (n == 1)
    {
        fmpz_poly_get_coeff_mpz(mpq_numref(rop), f, 0);
        mpz_set_si(mpq_denref(rop), 1);
    }
    else
    {
        n--;
        fmpz_poly_get_coeff_mpz(mpq_numref(rop), f, n);
        mpz_set_si(mpq_denref(rop), 1);
        mpq_init(temp);
        do {
            n--;
            mpq_mul(temp, rop, a);
            fmpz_poly_get_coeff_mpz(mpq_numref(rop), f, n);
            mpz_set_si(mpq_denref(rop), 1);
            mpq_add(rop, rop, temp);
        } while (n);
        mpq_clear(temp);
    }
}

/**
 * \ingroup  Evaluation
 * 
 * Evaluates the rational function \c f at the rational \c a.
 * 
 * If the denominator evaluates to zero at \c a, returns \c 0 and does not 
 * modify any of the variables.  Otherwise, returns \c 1 and sets \c rop to 
 * the rational \f$f(a)\f$.
 * 
 * In the special case where \c a is an integer, this method uses the method 
 * <tt>fmpz_poly_evaluate</tt> from the <tt>fmpz_poly</tt> module.  Otherwise, 
 * a simple implementation of Horner's method is employed.
 * 
 * \todo  The second argument should be declared \c const, but as of 
 *        FLINT 1.5.0 this generates a compile-time warning.
 */
int fmpz_poly_q_evaluate(mpq_t rop, fmpz_poly_q_ptr f, const mpq_t a)
{
    if (mpz_cmp_si(mpq_denref(a), 1))  // a is not an integer
    {
        mpq_t mpqnum, mpqden;

        mpq_init(mpqden);
        _fmpz_poly_evaluate_mpq_horner(mpqden, f->den, a);
        if (mpq_sgn(mpqden) == 0)
        {
            mpq_clear(mpqden);
            return 0;
        }
        
        mpq_init(mpqnum);
        _fmpz_poly_evaluate_mpq_horner(mpqnum, f->num, a);
        
        mpq_div(rop, mpqnum, mpqden);
        mpq_clear(mpqnum);
        mpq_clear(mpqden);
        mpq_canonicalize(rop);
        return 1;
    }
    else  // a is an integer
    {
        fmpz_t num, den, _a;
        ulong dnum = FLINT_MAX(fmpz_poly_length(f->num), 1);
        ulong dden = FLINT_MAX(fmpz_poly_length(f->den), 1);

        _a = fmpz_init(mpz_size(mpq_numref(a)));
        mpz_to_fmpz(_a, mpq_numref(a));

        num = fmpz_init(f->num->limbs + dnum + dnum * mpz_size(mpq_numref(a)));
        den = fmpz_init(f->den->limbs + dden + dden * mpz_size(mpq_numref(a)));

        fmpz_poly_evaluate(den, f->den, _a);
        if (fmpz_is_zero(den))
        {
            fmpz_clear(_a);
            fmpz_clear(num);
            fmpz_clear(den);
            return 0;
        }
        
        fmpz_poly_evaluate(num, f->num, _a);
        
        fmpz_to_mpz(mpq_numref(rop), num);
        fmpz_to_mpz(mpq_denref(rop), den);
        mpq_canonicalize(rop);
        fmpz_clear(_a);
        fmpz_clear(num);
        fmpz_clear(den);
        return 1;
    }
}

/**
 * \addtogroup StringConversions
 * 
 * The following three methods enable users to construct elements of type 
 * \c fmpz_poly_q_ptr from strings or to obtain string representations of such 
 * elements.
 *
 * The format used is based on the FLINT format for integer polynomials of 
 * type <tt>fmpz_poly_t</tt>, which we recall first: 
 * 
 * A non-zero polynomial \f$a_0 + a_1 X + \dotsb + a_n X^n\f$ of length 
 * \f$n+1\f$ is represented by the string <tt>n+1  a_0 a_1 ... a_n</tt>, 
 * where there are two space characters following the length and single space 
 * characters separating the individual coefficients.  There is no leading or 
 * trailing white-space.  In contrast, the zero polynomial is represented by 
 * <tt>0</tt>.
 * 
 * We adapt this notation for rational functions as follows.  We denote the 
 * zero function by <tt>0</tt>.  Given a non-zero function with numerator 
 * and denominator string representations \c num and \c den, respectively, 
 * we use the string <tt>num/den</tt> to represent the rational function, 
 * unless the denominator is equal to one, in which case we simply use \c num.
 * 
 * There is also a <tt>_pretty</tt> variant available, which bases the string 
 * parts for the numerator and denominator on the output of the function 
 * <tt>fmpz_poly_to_string_pretty</tt> and introduces parentheses where 
 * necessary.
 * 
 * Note that currently these functions are not optimized for performance and 
 * are intended to be used only for debugging purposes or one-off input and 
 * output, rather than as a low-level parser.
 */

/**
 * \ingroup  StringConversions
 * 
 * Sets the rational function \c rop to the value specified by the 
 * null-terminated string \c s.
 *
 * This method has now already been somewhat improved and is not very tolerant 
 * in the handling of malformed input.  It expects either legitimate input for 
 * an \c fmpz_poly_t element, or two such inputs separated by a <tt>/</tt> 
 * only, in which case it is also assumed that the second polynomial is 
 * non-zero.
 * 
 * The rational function is brought into canonical form by calling 
 * #fmpz_poly_q_canonicalize() in this function.
 * 
 * Returns \c 1 if the string represents a valid rational function and 
 * \c 0 otherwise.
 */
int fmpz_poly_q_from_string(fmpz_poly_q_ptr rop, const char *s)
{
    int ans, i, m;
    size_t len;
    char * numstr;
    
    len = strlen(s);
    
    for (m = 0; m < len; m++)
    {
        if (s[m] == '/')
            break;
    }
    
    if (m == len)
    {
        ans = fmpz_poly_from_string(rop->num, s);
        fmpz_poly_zero(rop->den);
        fmpz_poly_set_coeff_si(rop->den, 0, 1);
        return ans;
    }
    else
    {
        numstr = (char *) malloc((m+1) * sizeof(char));
        if (numstr == NULL)
        {
            printf("ERROR (fmpz_poly_q_from_string).  Memory allocation failed.\n");
            abort();
        }
        
        for (i = 0; i < m; i++)
            numstr[i] = s[i];
        numstr[i] = '\0';
        
        ans = fmpz_poly_from_string(rop->num, numstr);
        ans = fmpz_poly_from_string(rop->den, (char *) (s + (m+1))) & ans;
        if (ans)
            fmpz_poly_q_canonicalize(rop);
        else
            fmpz_poly_q_zero(rop);
        free(numstr);
        return ans;
    }
}

/**
 * \ingroup  StringConversions
 * 
 * Returns the string representation of the rational function \c op.
 */
char * fmpz_poly_q_to_string(const fmpz_poly_q_ptr op)
{
    int i, j;
    char * str;
    char * numstr;
    char * denstr;
    
    if (_fmpz_poly_is_one(op->den))
    {
        numstr = fmpz_poly_to_string(op->num);
        i = strlen(numstr)-1;
        if (numstr[i] == ' ')
        {
            numstr[i] = '\0';
        }
        return numstr;
    }
    
    numstr = fmpz_poly_to_string(op->num);
    denstr = fmpz_poly_to_string(op->den);
    
    i = strlen(numstr)-1;
    if (numstr[i] == ' ')
        numstr[i] = '\0';
    i = strlen(denstr)-1;
    if (denstr[i] == ' ')
        denstr[i] = '\0';
    
    str = (char *) malloc((strlen(numstr) + strlen(denstr) + 2) * sizeof(char));
    if (str == NULL)
    {
        printf("ERROR (fmpz_poly_q_to_string).  Memory allocation failed.\n");
        abort();
    }
    
    for (i = 0; i < strlen(numstr); i++)
        str[i] = numstr[i];
    str[i++] = '/';
    for (j = 0; j < strlen(denstr); j++)
        str[i++] = denstr[j];
    str[i] = '\0';
    
    free(numstr);
    free(denstr);
    
    return str;
}

/**
 * \ingroup  StringConversions
 * 
 * Returns the pretty string representation of \c op.
 * 
 * Returns the pretty string representation of the rational function \c op, 
 * using the string \c x as the variable name.
 */
char * fmpz_poly_q_to_string_pretty(const fmpz_poly_q_ptr op, const char *x)
{
    int i, j;
    char * str;
    char * numstr;
    char * denstr;
    
    if (_fmpz_poly_is_one(op->den))
    {
        return fmpz_poly_to_string_pretty(op->num, x);
    }
    
    numstr = fmpz_poly_to_string_pretty(op->num, x);
    denstr = fmpz_poly_to_string_pretty(op->den, x);
    
    str = (char *) malloc((strlen(numstr) + strlen(denstr) + 6) * sizeof(char));
    if (str == NULL)
    {
        printf("ERROR (fmpz_poly_q_to_string_pretty).  Memory allocation failed.\n");
        abort();
    }
    
    i = 0;
    if (fmpz_poly_degree(op->num) > 0)
    {
        str[i++] = '(';
        for (j = 0; j < strlen(numstr); j++)
            str[i++] = numstr[j];
        str[i++] = ')';
    }
    else
    {
        for (j = 0; j < strlen(numstr); j++)
            str[i++] = numstr[j];
    }
    str[i++] = '/';
    if (fmpz_poly_degree(op->den) > 0)
    {
        str[i++] = '(';
        for (j = 0; j < strlen(denstr); j++)
            str[i++] = denstr[j];
        str[i++] = ')';
    }
    else
    {
        for (j = 0; j < strlen(denstr); j++)
            str[i++] = denstr[j];
    }
    str[i] = '\0';
    
    free(numstr);
    free(denstr);
    
    return str;
}

/**
 * \ingroup  StringConversions
 * 
 * Prints the standard non-pretty representation of <tt>op</tt> 
 * to <tt>stdout</tt>.
 */
void fmpz_poly_q_print(const fmpz_poly_q_ptr op)
{
    char *str;

    str = fmpz_poly_q_to_string(op);
    printf("%s", str);
    free(str);
}

/**
 * \ingroup  StringConversions
 * 
 * Prints the standard pretty representation of <tt>op</tt> 
 * to <tt>stdout</tt>.
 */
void fmpz_poly_q_print_pretty(const fmpz_poly_q_ptr op, const char *x)
{
    char *str;

    str = fmpz_poly_q_to_string_pretty(op, x);
    printf("%s", str);
    free(str);
}

