///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2009, 2010 Sebastian Pancratz                               //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

#ifndef FMPZ_POLY_Q_H
#define FMPZ_POLY_Q_H

#include <stdio.h>
#include <gmp.h>

#include "fmpz.h"
#include "fmpz_poly.h"

/**
 * \file    fmpz_poly_q.h
 * \brief   Fast implementation of the rational function field
 * \author  Sebastian Pancratz
 * \date    Oct 2009 -- Dec 2010
 */

#ifndef __GNUC__
    #define __asm__     asm
    #define __inline__  inline
#endif

/**
 * \ingroup  Definitions
 * \brief    Data type for a rational function over the rationals or, 
 *           equivalently, the integers.
 * \details  We represent a rational function over \f$\mathbf{Q}\f$ as the 
 *           quotient of two coprime integer polynomials of type 
 *           <tt>fmpz_poly_t</tt>, enforcing that the leading coefficient of 
 *           the denominator is positive.  The zero function is represented 
 *           as \f$0/1\f$.
 */
typedef struct
{
    fmpz_poly_p num;  /*!< Numerator   */
    fmpz_poly_p den;  /*!< Denominator */
} fmpz_poly_q_struct;

/**
 * \ingroup  Definitions
 * \brief    Array of #fmpz_poly_q_struct of length one.
 * \details  This allows passing <em>by reference</em> without having to 
 *           explicitly allocate memory for the structure, as one would have 
 *           to when using pointers.
 */
typedef fmpz_poly_q_struct fmpz_poly_q_t[1];

/**
 * \ingroup  Definitions
 * \brief    Pointer to #fmpz_poly_q_struct.
 */
typedef fmpz_poly_q_struct *fmpz_poly_q_ptr;

/**
 * \ingroup  Definitions
 * \brief    Random state
 */
typedef gmp_randstate_t fmpz_poly_q_randstate_t;

/**
 * \ingroup  Definitions
 * \brief    Random state
 */
typedef __gmp_randstate_struct fmpz_poly_q_randstate_struct;

///////////////////////////////////////////////////////////////////////////////
// _fmpz_poly_*                                                              //
///////////////////////////////////////////////////////////////////////////////

/**
 * Returns whether the integer polynomials \c op is equal to one.
 */
static __inline__
int _fmpz_poly_is_one(const fmpz_poly_p op)
{
    return (op->length == 1) && fmpz_is_one(op->coeffs);
}

/**
 * Returns whether the integer polynomial \c op is equal to \f$\pm 1\f$.
 */
static __inline__
int _fmpz_poly_is_unit(const fmpz_poly_p op)
{
    return (op->length == 1) && (fmpz_is_one(op->coeffs) || fmpz_is_m1(op->coeffs));
}

///////////////////////////////////////////////////////////////////////////////
// fmpz_poly_q_*                                                             //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Accessing numerator and denominator

/**
 * \def      fmpz_poly_q_numref(op)
 * \brief    Returns a reference to the numerator of \c op.
 * \ingroup  NumDen
 * \details  The <tt>fmpz_poly</tt> functions can be used on the result of 
 *           this macro.
 */
#define fmpz_poly_q_numref(op)  ((op)->num)

/**
 * \def      fmpz_poly_q_denref(op)
 * \brief    Returns a reference to the denominator of \c op.
 * \ingroup  NumDen
 * \details  The <tt>fmpz_poly</tt> functions can be used on the result of 
 *           this macro.
 */
#define fmpz_poly_q_denref(op)  ((op)->den)

/**
 * \brief    Sets \c rop to the numerator of \c op.
 * \ingroup  NumDen
 * 
 * \param[out] rop  Integer polynomial
 * \param[in] op    Rational function
 */
static __inline__ 
void fmpz_poly_q_get_num(fmpz_poly_p rop, const fmpz_poly_q_ptr op)
{
    fmpz_poly_set(rop, op->num);
}

/**
 * \brief    Sets \c rop to the denominator of \c op.
 * \ingroup  NumDen
 * 
 * \param[out] rop  Integer polynomial
 * \param[in] op    Rational function
 */
static __inline__ 
void fmpz_poly_q_get_den(fmpz_poly_p rop, const fmpz_poly_q_ptr op)
{
    fmpz_poly_set(rop, op->den);
}

/**
 * \brief    Sets the numerator of \c rop to the polynomial \c op. 
 * \ingroup  NumDen
 * 
 * \param[out] rop  Integer polynomial
 * \param[in] op    Rational function
 */
static __inline__ 
void fmpz_poly_q_set_num(fmpz_poly_q_ptr rop, const fmpz_poly_p op)
{
    fmpz_poly_set(rop->num, op);
}

/**
 * \brief    Sets the denominator of \c rop to the polynomial \c op. 
 * \ingroup  NumDen
 * 
 * \param[out] rop  Integer polynomial
 * \param[in] op    Rational function
 */
static __inline__ 
void fmpz_poly_q_set_den(fmpz_poly_q_ptr rop, const fmpz_poly_p op)
{
    fmpz_poly_set(rop->den, op);
}

void fmpz_poly_q_canonicalize(fmpz_poly_q_ptr rop);

///////////////////////////////////////////////////////////////////////////////
// Memory management

/**
 * \ingroup  MemoryManagement
 * 
 * Initializes the element \c rop to zero.
 * 
 * This function should be called once for the #fmpz_poly_q_ptr \c rop, before 
 * using it with other <tt>fmpz_poly_q</tt> functions, or following a 
 * preceeding call to #fmpz_poly_q_clear().
 */
static __inline__ 
void fmpz_poly_q_init(fmpz_poly_q_ptr rop)
{
    rop->num = (fmpz_poly_p) malloc(sizeof(fmpz_poly_struct));
    rop->den = (fmpz_poly_p) malloc(sizeof(fmpz_poly_struct));
    fmpz_poly_init(rop->num);
    fmpz_poly_init(rop->den);
    fmpz_poly_set_coeff_si(rop->den, 0, 1);
}

/**
 * \ingroup  MemoryManagement
 *
 * Clears the element \c rop.
 * 
 * This function should only be called on an element which has been 
 * initialized.
 */
static __inline__ 
void fmpz_poly_q_clear(fmpz_poly_q_ptr rop)
{
    if (rop->num != NULL)
    {
        fmpz_poly_clear(rop->num);
        free(rop->num);
        rop->num = NULL;
    }
    if (rop->den != NULL);
    {
        fmpz_poly_clear(rop->den);
        free(rop->den);
        rop->den = NULL;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Randomisation

static 
void fmpz_poly_q_randinit(fmpz_poly_q_randstate_t state)
{
    gmp_randinit_default(state);
    gmp_randseed_ui(state, 123456789ul);
}

static 
void fmpz_poly_q_randclear(fmpz_poly_q_randstate_t state)
{
    gmp_randclear(state);
}

void fmpz_randtest(fmpz_t * f, fmpz_poly_q_randstate_t R, ulong bits);

void fmpz_randtest_not_zero(fmpz_t * f, fmpz_poly_q_randstate_t R, ulong bits);

void fmpz_poly_randtest(fmpz_poly_t poly, 
                             fmpz_poly_q_randstate_t R, ulong len, ulong bits);

void fmpz_poly_randtest_not_zero(fmpz_poly_t poly, 
                             fmpz_poly_q_randstate_t R, ulong len, ulong bits);

void fmpz_poly_q_randtest(fmpz_poly_q_t poly, fmpz_poly_q_randstate_t R, 
                             ulong len1, ulong bits1, ulong len2, ulong bits2);

void 
fmpz_poly_q_randtest_not_zero(fmpz_poly_q_t poly, fmpz_poly_q_randstate_t R, 
                             ulong len1, ulong bits1, ulong len2, ulong bits2);

///////////////////////////////////////////////////////////////////////////////
// Assignment and basic manipulation

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to the same value as the element \c op.
 */
static __inline__ 
void fmpz_poly_q_set(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op)
{
    if (rop != op)
    {
        fmpz_poly_set(rop->num, op->num);
        fmpz_poly_set(rop->den, op->den);
    }
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to the value given by the \c int \c op.
 */
static __inline__ 
void fmpz_poly_q_set_si(fmpz_poly_q_ptr rop, long op)
{
    fmpz_poly_zero(rop->num);
    fmpz_poly_zero(rop->den);
    fmpz_poly_set_coeff_si(rop->num, 0, op);
    fmpz_poly_set_coeff_si(rop->den, 0, 1);
}

/**
 * \ingroup  Assignment
 * 
 * Swaps the elements \c op1 and \c op2.
 * 
 * This is done efficiently by swapping pointers.
 */
static __inline__ 
void fmpz_poly_q_swap(fmpz_poly_q_ptr op1, fmpz_poly_q_ptr op2)
{
    if (op1 != op2)
    {
        fmpz_poly_swap(op1->num, op2->num);
        fmpz_poly_swap(op1->den, op2->den);
    }
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to zero.
 */
static __inline__
void fmpz_poly_q_zero(fmpz_poly_q_ptr rop)
{
    fmpz_poly_zero(rop->num);
    fmpz_poly_zero(rop->den);
    fmpz_poly_set_coeff_si(rop->den, 0, 1);
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to one.
 */
static __inline__ 
void fmpz_poly_q_one(fmpz_poly_q_ptr rop)
{
    fmpz_poly_zero(rop->num);
    fmpz_poly_zero(rop->den);
    fmpz_poly_set_coeff_si(rop->num, 0, 1);
    fmpz_poly_set_coeff_si(rop->den, 0, 1);
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to the additive inverse of \c op.
 */
static __inline__ 
void fmpz_poly_q_neg(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op)
{
    fmpz_poly_neg(rop->num, op->num);
    fmpz_poly_set(rop->den, op->den);
}

/**
 * \ingroup  Assignment
 *
 * Sets the element \c rop to the multiplicative inverse of \c op.
 * 
 * Assumes that the element \c op is non-zero.
 */
static __inline__ 
void fmpz_poly_q_inv(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op)
{
    // Assertion!
    if (fmpz_poly_degree(op->num) < 0)
    {
        printf("ERROR (fmpz_poly_q_canonicalize).  Denominator is zero.\n");
        abort();
    }
    
    if (rop == op)
    {
        fmpz_poly_swap(rop->num, rop->den);
        if (fmpz_sgn(fmpz_poly_lead(rop->den)) < 0)
        {
            fmpz_poly_neg(rop->num, rop->num);
            fmpz_poly_neg(rop->den, rop->den);
        }
    }
    else
    {
        if (fmpz_sgn(fmpz_poly_lead(op->num)) > 0)
        {
            fmpz_poly_set(rop->num, op->den);
            fmpz_poly_set(rop->den, op->num);
        }
        else
        {
            fmpz_poly_neg(rop->num, op->den);
            fmpz_poly_neg(rop->den, op->num);
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// Comparison

/**
 * \ingroup  Comparison
 * 
 * Returns whether the element \c op is zero.
 */
static __inline__ 
int fmpz_poly_q_is_zero(const fmpz_poly_q_ptr op)
{
    return fmpz_poly_degree(op->num) < 0;
}

/**
 * \ingroup  Comparison
 * 
 * Returns whether the element \c rop is equal to the constant polynomial 
 * \f$1\f$.
 */
static __inline__ 
int fmpz_poly_q_is_one(const fmpz_poly_q_ptr op)
{
    return _fmpz_poly_is_one(op->num) && _fmpz_poly_is_one(op->den);
}

/**
 * \ingroup  Comparison
 * 
 * Returns whether the two elements \c op1 and \c op2 are equal.
 */
static __inline__ 
int fmpz_poly_q_equal(const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2)
{
    return fmpz_poly_equal(op1->num, op2->num) && fmpz_poly_equal(op1->den, op2->den);
}

///////////////////////////////////////////////////////////////////////////////
// Addition/ subtraction

void fmpz_poly_q_add(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2);
void fmpz_poly_q_sub(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2);

void fmpz_poly_q_addmul(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2);
void fmpz_poly_q_submul(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2);

///////////////////////////////////////////////////////////////////////////////
// Scalar multiplication and division

void fmpz_poly_q_scalar_mul_si(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, long x);
void fmpz_poly_q_scalar_mul_mpz(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, const mpz_t x);
void fmpz_poly_q_scalar_mul_mpq(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, const mpq_t x);

void fmpz_poly_q_scalar_div_si(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, long x);
void fmpz_poly_q_scalar_div_mpz(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, const mpz_t x);
void fmpz_poly_q_scalar_div_mpq(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, const mpq_t x);

///////////////////////////////////////////////////////////////////////////////
// Multiplication

void fmpz_poly_q_mul(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2);

///////////////////////////////////////////////////////////////////////////////
// Division

void fmpz_poly_q_div(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op1, const fmpz_poly_q_ptr op2);

///////////////////////////////////////////////////////////////////////////////
// Powering

void fmpz_poly_q_power(fmpz_poly_q_ptr rop, const fmpz_poly_q_ptr op, unsigned long exp);

///////////////////////////////////////////////////////////////////////////////
// Derivative

void fmpz_poly_q_derivative(fmpz_poly_q_ptr rop, fmpz_poly_q_ptr op);

///////////////////////////////////////////////////////////////////////////////
// Evaluation

int fmpz_poly_q_evaluate(mpq_t rop, fmpz_poly_q_ptr f, const mpq_t a);

///////////////////////////////////////////////////////////////////////////////
// String conversion

int fmpz_poly_q_from_string(fmpz_poly_q_ptr rop, const char *s);

char * fmpz_poly_q_to_string(const fmpz_poly_q_ptr op);
char * fmpz_poly_q_to_string_pretty(const fmpz_poly_q_ptr op, const char *x);

void fmpz_poly_q_print(const fmpz_poly_q_ptr op);
void fmpz_poly_q_print_pretty(const fmpz_poly_q_ptr op, const char *x);

#endif

