///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2010 Sebastian Pancratz                                     //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
    extern "C" {
#endif

#include "fmpz_poly_q.h"

int main(int argc, char *argv[])
{
    fmpz_poly_q_randstate_t R;
    ulong i, N = 1000;
    
    fmpz_poly_q_randinit(R);
    srand(123);

    printf("derivative..."); fflush(stdout);
    
    /* Check aliasing */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        
        fmpz_poly_q_derivative(g, f);
        fmpz_poly_q_derivative(f, f);
        
        result = fmpz_poly_q_equal(f, g);
        if (!result)
        {
            char * strf, * strg;
            strf = fmpz_poly_q_to_string(f);
            strg = fmpz_poly_q_to_string(g);
            printf("FAIL\n");
            printf("  Aliasing\n");
            printf("  f = %s\n", strf);
            printf("  g = %s\n", strg);
            free(strf);
            free(strg);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
    }

    /* d(f) == 0 for constant f */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f;
        ulong len = 1, bits = 100;
        
        fmpz_poly_q_init(f);
        
        fmpz_poly_q_randtest(f, R, len, rand() % bits, len, rand () % bits);
        
        fmpz_poly_q_derivative(f, f);
        
        result = fmpz_poly_q_is_zero(f);
        if (!result)
        {
            char * strf;
            strf = fmpz_poly_q_to_string(f);
            printf("FAIL\n");
            printf("  d(f) == 0 for constant f\n");
            printf("  f = %s\n", strf);
            free(strf);
            abort();
        }
        
        fmpz_poly_q_clear(f);
    }

    /* d(f + g) == df + dg */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t a, b, f, g, der;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(a);
        fmpz_poly_q_init(b);
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(der);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        
        fmpz_poly_q_set(a, f);
        fmpz_poly_q_set(b, g);
        
        fmpz_poly_q_add(der, f, g);
        fmpz_poly_q_derivative(der, der);
        fmpz_poly_q_derivative(f, f);
        fmpz_poly_q_derivative(g, g);
        fmpz_poly_q_add(f, f, g);
        
        result = fmpz_poly_q_equal(der, f);
        if (!result)
        {
            char *strf, *strg;
            strf = fmpz_poly_q_to_string(der);
            strg = fmpz_poly_q_to_string(f);
            printf("FAIL\n");
            printf("  d(f + g) == df + dg\n");
            printf("  f = "), fmpz_poly_q_print(a), printf("\n");
            printf("  g = "), fmpz_poly_q_print(b), printf("\n");
            printf("  lhs = %s\n", strf);
            printf("  rhs = %s\n", strg);
            free(strf);
            free(strg);
            abort();
        }
        
        fmpz_poly_q_clear(a);
        fmpz_poly_q_clear(b);
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(der);
    }

    /* d(a f) == a df for rational a */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g;
        fmpz_t num, den;
        mpq_t x;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        num = fmpz_init(bits / FLINT_BITS + 2);
        den = fmpz_init(bits / FLINT_BITS + 2);
        mpq_init(x);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_randtest(&num, R, rand() % bits);
        fmpz_randtest_not_zero(&den, R, rand() % bits);
        fmpz_to_mpz(mpq_numref(x), num);
        fmpz_to_mpz(mpq_denref(x), den);
        mpq_canonicalize(x);
        
        fmpz_poly_q_scalar_mul_mpq(g, f, x);
        fmpz_poly_q_derivative(g, g);
        fmpz_poly_q_derivative(f, f);
        fmpz_poly_q_scalar_mul_mpq(f, f, x);
        
        result = fmpz_poly_q_equal(g, f);
        if (!result)
        {
            char * strf, * strg;
            strf = fmpz_poly_q_to_string(g);
            strg = fmpz_poly_q_to_string(f);
            printf("FAIL\n");
            printf("  d(a f) == a df for rational a\n");
            printf("  lhs = %s\n", strf);
            printf("  rhs = %s\n", strg);
            free(strf);
            free(strg);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_clear(num);
        fmpz_clear(den);
        mpq_clear(x);
    }

    /* d(f g) == df g + f dg */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t lhs, rhs, f, g;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(lhs);
        fmpz_poly_q_init(rhs);
        
        fmpz_poly_q_randtest(f, R, len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, len, rand() % bits, rand() % len, rand() % bits);
        
        fmpz_poly_q_mul(lhs, f, g);
        fmpz_poly_q_derivative(lhs, lhs);
        fmpz_poly_q_derivative(rhs, f);
        fmpz_poly_q_mul(rhs, rhs, g);
        fmpz_poly_q_derivative(g, g);
        fmpz_poly_q_mul(f, f, g);
        fmpz_poly_q_add(rhs, rhs, f);
        
        result = fmpz_poly_q_equal(lhs, rhs);
        if (!result)
        {
            char * str1, * str2;
            str1 = fmpz_poly_q_to_string(lhs);
            str2 = fmpz_poly_q_to_string(rhs);
            printf("FAIL\n");
            printf("  d(f g) == df g + f dg\n");
            printf("  lhs = %s\n", str1);
            printf("  rhs = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(lhs);
        fmpz_poly_q_clear(rhs);
    }
    
    printf("PASS\n");
    
    fmpz_poly_q_randclear(R);
    flint_stack_cleanup();
    return EXIT_SUCCESS;
}

#ifdef __cplusplus
    }
#endif

