///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2010 Sebastian Pancratz                                     //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
    extern "C" {
#endif

#include "fmpz_poly_q.h"

int main(int argc, char *argv[])
{
    fmpz_poly_q_randstate_t R;
    ulong i, N = 1000;
    
    fmpz_poly_q_randinit(R);
    srand(123);
    
    printf("evaluate..."); fflush(stdout);

    /* Aliasing */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f;
        fmpz_t anum, aden;
        mpq_t a, x;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        anum = fmpz_init(bits / FLINT_BITS + 2);
        aden = fmpz_init(bits / FLINT_BITS + 2);
        mpq_init(a);
        mpq_init(x);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_randtest(&anum, R, rand() % bits);
        fmpz_randtest_not_zero(&aden, R, rand() % bits);
        fmpz_to_mpz(mpq_numref(a), anum);
        fmpz_to_mpz(mpq_denref(a), aden);
        mpq_canonicalize(a);
        
        fmpz_poly_q_evaluate(x, f, a);
        fmpz_poly_q_evaluate(a, f, a);
        
        result = mpq_equal(x, a);
        if (!result)
        {
            char * str1, * str2;
            str1 = mpq_get_str(NULL, 10, x);
            str2 = mpq_get_str(NULL, 10, a);
            printf("FAIL\n");
            printf("  Aliasing\n");
            printf("  x = %s\n", str1);
            printf("  a = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_clear(anum);
        fmpz_clear(aden);
        mpq_clear(a);
        mpq_clear(x);
    }
    
    /* Checking (f + g)(a) == f(a) + g(a) */
    for (i = 0; i < N; i++)
    {
        int ansf, ansg, anssum, result;
        fmpz_poly_q_t f, g, sum;
        fmpz_t anum, aden;
        mpq_t a;
        mpq_t x, y, s, z;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(sum);
        anum = fmpz_init(bits / FLINT_BITS + 2);
        aden = fmpz_init(bits / FLINT_BITS + 2);
        mpq_init(a);
        mpq_init(x);
        mpq_init(y);
        mpq_init(s);
        mpq_init(z);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_randtest(&anum, R, rand() % bits);
        fmpz_randtest_not_zero(&aden, R, rand() % bits);
        fmpz_to_mpz(mpq_numref(a), anum);
        fmpz_to_mpz(mpq_denref(a), aden);
        mpq_canonicalize(a);
        
        ansf = fmpz_poly_q_evaluate(x, f, a);
        ansg = fmpz_poly_q_evaluate(y, g, a);
        mpq_add(s, x, y);
        fmpz_poly_q_add(sum, f, g);
        anssum = fmpz_poly_q_evaluate(z, sum, a);
        
        result = (mpq_equal(s, z) || ansf * ansg * anssum == 0);
        if (!result)
        {
            printf("FAIL\n");
            printf("  (f + g)(a) == f(a) + g(a)\n");
            printf("  f = "), fmpz_poly_q_print(f); printf("\n");
            printf("  g = "), fmpz_poly_q_print(g); printf("\n");
            printf("  f + g = "), fmpz_poly_q_print(sum); printf("\n");
            gmp_printf("  a = %Qd\n", a);
            gmp_printf("  f(a) = %Qd\n", x);
            gmp_printf("  g(a) = %Qd\n", y);
            gmp_printf("  f(a) + g(a) = %Qd\n", y);
            gmp_printf("  (f + g)(a) = %Qd\n", z);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(sum);
        fmpz_clear(anum);
        fmpz_clear(aden);
        mpq_clear(a);
        mpq_clear(x);
        mpq_clear(y);
        mpq_clear(s);
        mpq_clear(z);
    }
    printf("PASS\n");
    
    fmpz_poly_q_randclear(R);
    flint_stack_cleanup();
    return EXIT_SUCCESS;
}

#ifdef __cplusplus
    }
#endif

