///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2010 Sebastian Pancratz                                     //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
    extern "C" {
#endif

#include "fmpz_poly_q.h"

int main(int argc, char *argv[])
{
    fmpz_poly_q_randstate_t R;
    ulong i, N = 1000;
    
    fmpz_poly_q_randinit(R);
    srand(123);
    
    printf("mul..."); fflush(stdout);
    
    /* Aliasing f and h in h = f * g */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g, h;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(h);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        
        fmpz_poly_q_mul(h, f, g);
        fmpz_poly_q_mul(f, f, g);
        
        result = fmpz_poly_q_equal(f, h);
        if (!result)
        {
            char * strf, * strh;
            strf = fmpz_poly_q_to_string(f);
            strh = fmpz_poly_q_to_string(h);
            printf("FAIL\n");
            printf("  f = %s\n", strf);
            printf("  h = %s\n", strh);
            free(strf);
            free(strh);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(h);
    }
    
    /* Aliasing g and h in h = f * g */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g, h;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(h);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        
        fmpz_poly_q_mul(h, f, g);
        fmpz_poly_q_mul(g, f, g);
        
        result = fmpz_poly_q_equal(g, h);
        if (!result)
        {
            char * strg, * strh;
            strg = fmpz_poly_q_to_string(g);
            strh = fmpz_poly_q_to_string(h);
            printf("FAIL\n");
            printf("  g = %s\n", strg);
            printf("  h = %s\n", strh);
            free(strg);
            free(strh);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(h);
    }

    /* Aliasing f and g in h = f * g */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, h;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(h);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        
        fmpz_poly_q_mul(h, f, f);
        fmpz_poly_q_mul(f, f, f);
        
        result = fmpz_poly_q_equal(f, h);
        if (!result)
        {
            char * strf, * strh;
            strf = fmpz_poly_q_to_string(f);
            strh = fmpz_poly_q_to_string(h);
            printf("FAIL\n");
            printf("  f = %s\n", strf);
            printf("  h = %s\n", strh);
            free(strf);
            free(strh);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(h);
    }
    
    /* Verify commutativity */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g, h;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(h);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        
        fmpz_poly_q_mul(h, f, g);
        fmpz_poly_q_mul(f, g, f);
        
        result = fmpz_poly_q_equal(f, h);
        if (!result)
        {
            char * strf, * strh;
            strf = fmpz_poly_q_to_string(f);
            strh = fmpz_poly_q_to_string(h);
            printf("FAIL\n");
            printf("  f = %s\n", strf);
            printf("  h = %s\n", strh);
            free(strf);
            free(strh);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(h);
    }

    /* Verify (f + g) * h = f * h + g * h */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g, h, lhs, rhs;
        ulong len = 50, bits = 50;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(h);
        fmpz_poly_q_init(lhs);
        fmpz_poly_q_init(rhs);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(h, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        
        fmpz_poly_q_add(lhs, f, g);
        fmpz_poly_q_mul(lhs, lhs, h);
        
        fmpz_poly_q_mul(f, f, h);
        fmpz_poly_q_mul(g, g, h);
        fmpz_poly_q_add(rhs, f, g);
        
        result = fmpz_poly_q_equal(lhs, rhs);
        if (!result)
        {
            char * str1, * str2;
            str1 = fmpz_poly_q_to_string(lhs);
            str2 = fmpz_poly_q_to_string(rhs);
            printf("FAIL\n");
            printf("  lhs = %s\n", str1);
            printf("  rhs = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(h);
        fmpz_poly_q_clear(lhs);
        fmpz_poly_q_clear(rhs);
    }
    
    printf("PASS\n");
    
    fmpz_poly_q_randclear(R);
    flint_stack_cleanup();
    return EXIT_SUCCESS;
}

#ifdef __cplusplus
    }
#endif

