///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2010 Sebastian Pancratz                                     //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
    extern "C" {
#endif

#include "fmpz_poly_q.h"

#include <math.h>

int main(int argc, char *argv[])
{
    fmpz_poly_q_randstate_t R;
    ulong i, N = 1000;
    
    fmpz_poly_q_randinit(R);
    srand(123);
    
    printf("scalar_div_mpz..."); fflush(stdout);
    
    /* Aliasing f and g in g = f / c */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g;
        ulong len = 50, bits = 50;
        mpz_t c;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        mpz_init(c);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        while (!mpz_sgn(c))
            mpz_rrandomb(c, R, rand() % bits);
        if (rand() % 2)
            mpz_neg(c, c);
        
        fmpz_poly_q_scalar_div_mpz(g, f, c);
        fmpz_poly_q_scalar_div_mpz(f, f, c);
        
        result = fmpz_poly_q_equal(f, g);
        if (!result)
        {
            char * strf, * strg;
            strf = fmpz_poly_q_to_string(f);
            strg = fmpz_poly_q_to_string(g);
            printf("FAIL\n");
            printf("  f = %s\n", strf);
            printf("  g = %s\n", strg);
            free(strf);
            free(strg);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        mpz_clear(c);
    }
    
    /* (f + g) / c == f / c + g / c */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g, lhs, rhs;
        ulong len = 50, bits = 50;
        mpz_t c;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(lhs);
        fmpz_poly_q_init(rhs);
        mpz_init(c);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        while (!mpz_sgn(c))
            mpz_rrandomb(c, R, rand() % bits);
        if (rand() % 2)
            mpz_neg(c, c);
        
        fmpz_poly_q_add(lhs, f, g);
        fmpz_poly_q_scalar_div_mpz(lhs, lhs, c);

        fmpz_poly_q_scalar_div_mpz(f, f, c);
        fmpz_poly_q_scalar_div_mpz(g, g, c);
        fmpz_poly_q_add(rhs, f, g);
        
        result = fmpz_poly_q_equal(lhs, rhs);
        if (!result)
        {
            char * str1, * str2;
            str1 = fmpz_poly_q_to_string(lhs);
            str2 = fmpz_poly_q_to_string(rhs);
            printf("FAIL\n");
            printf("  lhs = %s\n", str1);
            printf("  rhs = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(lhs);
        fmpz_poly_q_clear(rhs);
        mpz_clear(c);
    }

    /* (f / c) / d == f / (c d) */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, lhs, rhs;
        ulong len = 50, bits = 50;
        mpz_t c, d;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(lhs);
        fmpz_poly_q_init(rhs);
        mpz_init(c);
        mpz_init(d);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        while (!mpz_sgn(c))
            mpz_rrandomb(c, R, rand() % bits);
        if (rand() % 2)
            mpz_neg(c, c);
        while (!mpz_sgn(d))
            mpz_rrandomb(d, R, rand() % bits);
        if (rand() % 2)
            mpz_neg(d, d);
        
        fmpz_poly_q_scalar_div_mpz(lhs, f, c);
        fmpz_poly_q_scalar_div_mpz(lhs, lhs, d);

        mpz_mul(c, c, d);
        fmpz_poly_q_scalar_div_mpz(rhs, f, c);
        
        result = fmpz_poly_q_equal(lhs, rhs);
        if (!result)
        {
            char * str1, * str2;
            str1 = fmpz_poly_q_to_string(lhs);
            str2 = fmpz_poly_q_to_string(rhs);
            printf("FAIL\n");
            printf("  lhs = %s\n", str1);
            printf("  rhs = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(lhs);
        fmpz_poly_q_clear(rhs);
        mpz_clear(c);
        mpz_clear(d);
    }

    /* (f / c) * c == 1 */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g;
        ulong len = 50, bits = 50;
        mpz_t c;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        mpz_init(c);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        while (!mpz_sgn(c))
            mpz_rrandomb(c, R, rand() % bits);
        if (rand() % 2)
            mpz_neg(c, c);
        
        fmpz_poly_q_set(g, f);
        fmpz_poly_q_scalar_div_mpz(f, f, c);
        fmpz_poly_q_scalar_mul_mpz(f, f, c);

        result = fmpz_poly_q_equal(f, g);
        if (!result)
        {
            char * str1, * str2;
            str1 = fmpz_poly_q_to_string(f);
            str2 = fmpz_poly_q_to_string(g);
            printf("FAIL\n");
            printf("  f = %s\n", str1);
            printf("  g = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        mpz_clear(c);
    }

    printf("PASS\n");
    
    fmpz_poly_q_randclear(R);
    flint_stack_cleanup();
    return EXIT_SUCCESS;
}

#ifdef __cplusplus
    }
#endif

