///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2010 Sebastian Pancratz                                     //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
    extern "C" {
#endif

#include "fmpz_poly_q.h"

#include <math.h>

int main(int argc, char *argv[])
{
    fmpz_poly_q_randstate_t R;
    ulong i, N = 1000;
    
    fmpz_poly_q_randinit(R);
    srand(123);
    
    printf("scalar_mul_mpz..."); fflush(stdout);
    
    /* Aliasing f and g in g = c * f */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g;
        ulong len = 50, bits = 50;
        fmpz_t c;
        mpz_t mc;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        c = fmpz_init(bits / FLINT_BITS + 2);
        mpz_init(mc);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_randtest(&c, R, rand() % bits);
        fmpz_to_mpz(mc, c);
        
        fmpz_poly_q_scalar_mul_mpz(g, f, mc);
        fmpz_poly_q_scalar_mul_mpz(f, f, mc);
        
        result = fmpz_poly_q_equal(f, g);
        if (!result)
        {
            char * strf, * strg;
            strf = fmpz_poly_q_to_string(f);
            strg = fmpz_poly_q_to_string(g);
            printf("FAIL\n");
            printf("  f = %s\n", strf);
            printf("  g = %s\n", strg);
            free(strf);
            free(strg);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_clear(c);
        mpz_clear(mc);
    }
    
    /* c (f + g) == c f + c g */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g, lhs, rhs;
        ulong len = 50, bits = 50;
        fmpz_t c;
        mpz_t mc;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(lhs);
        fmpz_poly_q_init(rhs);
        c = fmpz_init(bits / FLINT_BITS + 2);
        mpz_init(mc);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_randtest(&c, R, rand() % bits);
        fmpz_to_mpz(mc, c);
        
        fmpz_poly_q_add(lhs, f, g);
        fmpz_poly_q_scalar_mul_mpz(lhs, lhs, mc);

        fmpz_poly_q_scalar_mul_mpz(f, f, mc);
        fmpz_poly_q_scalar_mul_mpz(g, g, mc);
        fmpz_poly_q_add(rhs, f, g);
        
        result = fmpz_poly_q_equal(lhs, rhs);
        if (!result)
        {
            char * str1, * str2;
            str1 = fmpz_poly_q_to_string(lhs);
            str2 = fmpz_poly_q_to_string(rhs);
            printf("FAIL\n");
            printf("  lhs = %s\n", str1);
            printf("  rhs = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(lhs);
        fmpz_poly_q_clear(rhs);
        fmpz_clear(c);
        mpz_clear(mc);
    }

    /* d (c f) == (c d) f */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, lhs, rhs;
        ulong len = 50, bits = 50;
        fmpz_t c, d;
        mpz_t mc, md, mx;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(lhs);
        fmpz_poly_q_init(rhs);
        c = fmpz_init(bits / FLINT_BITS + 2);
        d = fmpz_init(bits / FLINT_BITS + 2);
        mpz_init(mc);
        mpz_init(md);
        mpz_init(mx);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_randtest(&c, R, rand() % bits);
        fmpz_randtest(&d, R, rand() % bits);
        fmpz_to_mpz(mc, c);
        fmpz_to_mpz(md, d);
        
        fmpz_poly_q_scalar_mul_mpz(lhs, f, mc);
        fmpz_poly_q_scalar_mul_mpz(lhs, lhs, md);

        mpz_mul(mx, mc, md);
        fmpz_poly_q_scalar_mul_mpz(rhs, f, mx);
        
        result = fmpz_poly_q_equal(lhs, rhs);
        if (!result)
        {
            char * str1, * str2;
            str1 = fmpz_poly_q_to_string(lhs);
            str2 = fmpz_poly_q_to_string(rhs);
            printf("FAIL\n");
            printf("  lhs = %s\n", str1);
            printf("  rhs = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(lhs);
        fmpz_poly_q_clear(rhs);
        fmpz_clear(c);
        fmpz_clear(d);
        mpz_clear(mc);
        mpz_clear(md);
        mpz_clear(mx);
    }

    printf("PASS\n");
    
    fmpz_poly_q_randclear(R);
    flint_stack_cleanup();
    return EXIT_SUCCESS;
}

#ifdef __cplusplus
    }
#endif

