///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2010 Sebastian Pancratz                                     //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
    extern "C" {
#endif

#include "fmpz_poly_q.h"

#include <math.h>

int main(int argc, char *argv[])
{
    fmpz_poly_q_randstate_t R;
    ulong i, N = 1000;
    
    fmpz_poly_q_randinit(R);
    srand(123);
    
    printf("scalar_mul_si..."); fflush(stdout);
    
    /* Aliasing f and g in g = c * f */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g;
        ulong len = 50, bits = 50;
        long c;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        c = rand() % LONG_MAX;
        if (rand() % 2)
            c = -c;
        
        fmpz_poly_q_scalar_mul_si(g, f, c);
        fmpz_poly_q_scalar_mul_si(f, f, c);
        
        result = fmpz_poly_q_equal(f, g);
        if (!result)
        {
            char * strf, * strg;
            strf = fmpz_poly_q_to_string(f);
            strg = fmpz_poly_q_to_string(g);
            printf("FAIL\n");
            printf("  f = %s\n", strf);
            printf("  g = %s\n", strg);
            free(strf);
            free(strg);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
    }
    
    /* c (f + g) == c f + c g */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, g, lhs, rhs;
        ulong len = 50, bits = 50;
        long c;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(g);
        fmpz_poly_q_init(lhs);
        fmpz_poly_q_init(rhs);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        fmpz_poly_q_randtest(g, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        c = rand() % LONG_MAX;
        if (rand() % 2)
            c = -c;
        
        fmpz_poly_q_add(lhs, f, g);
        fmpz_poly_q_scalar_mul_si(lhs, lhs, c);

        fmpz_poly_q_scalar_mul_si(f, f, c);
        fmpz_poly_q_scalar_mul_si(g, g, c);
        fmpz_poly_q_add(rhs, f, g);
        
        result = fmpz_poly_q_equal(lhs, rhs);
        if (!result)
        {
            char * str1, * str2;
            str1 = fmpz_poly_q_to_string(lhs);
            str2 = fmpz_poly_q_to_string(rhs);
            printf("FAIL\n");
            printf("  lhs = %s\n", str1);
            printf("  rhs = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(g);
        fmpz_poly_q_clear(lhs);
        fmpz_poly_q_clear(rhs);
    }

    /* d (c f) == (c d) f */
    for (i = 0; i < N; i++)
    {
        int result;
        fmpz_poly_q_t f, lhs, rhs;
        ulong len = 50, bits = 50;
        long c, d;
        
        fmpz_poly_q_init(f);
        fmpz_poly_q_init(lhs);
        fmpz_poly_q_init(rhs);
        
        fmpz_poly_q_randtest(f, R, rand() % len, rand() % bits, rand() % len, rand() % bits);
        c = rand() % ((long) sqrt((double) LONG_MAX));
        if (rand() % 2)
            c = -c;
        d = rand() % ((long) sqrt((double) LONG_MAX));
        if (rand() % 2)
            d = -d;
        
        fmpz_poly_q_scalar_mul_si(lhs, f, c);
        fmpz_poly_q_scalar_mul_si(lhs, lhs, d);

        fmpz_poly_q_scalar_mul_si(rhs, f, c * d);
        
        result = fmpz_poly_q_equal(lhs, rhs);
        if (!result)
        {
            char * str1, * str2;
            str1 = fmpz_poly_q_to_string(lhs);
            str2 = fmpz_poly_q_to_string(rhs);
            printf("FAIL\n");
            printf("  lhs = %s\n", str1);
            printf("  rhs = %s\n", str2);
            free(str1);
            free(str2);
            abort();
        }
        
        fmpz_poly_q_clear(f);
        fmpz_poly_q_clear(lhs);
        fmpz_poly_q_clear(rhs);
    }

    printf("PASS\n");
    
    fmpz_poly_q_randclear(R);
    flint_stack_cleanup();
    return EXIT_SUCCESS;
}

#ifdef __cplusplus
    }
#endif

